from typing import Any

import numpy as np
import pandas as pd
from tqdm.auto import tqdm

from qubx import QubxLogConfig, logger
from qubx.backtester.sentinels import NoDataContinue
from qubx.backtester.simulated_data import IterableSimulationData
from qubx.core.account import CompositeAccountProcessor
from qubx.core.basics import SW, DataType, Instrument, TransactionCostsCalculator
from qubx.core.context import StrategyContext
from qubx.core.exceptions import SimulationConfigError, SimulationError
from qubx.core.helpers import extract_parameters_from_object, full_qualified_class_name
from qubx.core.initializer import BasicStrategyInitializer
from qubx.core.interfaces import (
    CtrlChannel,
    IDataProvider,
    IMetricEmitter,
    IStrategy,
    IStrategyContext,
    ITimeProvider,
    StrategyState,
)
from qubx.core.loggers import StrategyLogging
from qubx.core.lookups import lookup
from qubx.data.helpers import CachedPrefetchReader
from qubx.loggers.inmemory import InMemoryLogsWriter
from qubx.pandaz.utils import _frame_to_str
from qubx.utils.time import now_ns

from .account import SimulatedAccountProcessor
from .broker import SimulatedBroker
from .data import SimulatedDataProvider
from .simulated_exchange import get_simulated_exchange
from .utils import (
    SetupTypes,
    SignalsProxy,
    SimulatedCtrlChannel,
    SimulatedScheduler,
    SimulatedTimeProvider,
    SimulationDataConfig,
    SimulationSetup,
    TimeGuardedWrapper,
)


class SimulationRunner:
    """
    A wrapper around the StrategyContext that encapsulates the simulation logic.
    This class is responsible for running a backtest context from a start time to an end time.
    """

    setup: SimulationSetup
    data_config: SimulationDataConfig
    start: pd.Timestamp
    stop: pd.Timestamp
    account_id: str
    portfolio_log_freq: str
    ctx: IStrategyContext
    logs_writer: InMemoryLogsWriter

    account: CompositeAccountProcessor
    channel: CtrlChannel
    time_provider: SimulatedTimeProvider
    scheduler: SimulatedScheduler
    strategy_params: dict[str, Any]
    strategy_class: str

    # adjusted times
    _stop: pd.Timestamp | None = None

    _data_source: IterableSimulationData
    _data_providers: list[SimulatedDataProvider]
    _exchange_to_data_provider: dict[str, SimulatedDataProvider]

    def __init__(
        self,
        setup: SimulationSetup,
        data_config: SimulationDataConfig,
        start: pd.Timestamp | str,
        stop: pd.Timestamp | str,
        account_id: str = "SimulatedAccount",
        portfolio_log_freq: str = "5Min",
        emitter: IMetricEmitter | None = None,
        strategy_state: StrategyState | None = None,
        initializer: BasicStrategyInitializer | None = None,
        warmup_mode: bool = False,
    ):
        """
        Initialize the BacktestContextRunner with a strategy context.

        Args:
            setup (SimulationSetup): The setup to run.
            data_config (SimulationDataConfig): The data setup to use.
            start (pd.Timestamp): The start time of the simulation.
            stop (pd.Timestamp): The end time of the simulation.
            account_id (str): The account id to use.
            portfolio_log_freq (str): The portfolio log frequency to use.
            emitter (IMetricEmitter): The emitter to use.
        """
        self.setup = setup
        self.data_config = data_config
        self.start = pd.Timestamp(start)
        self.stop = pd.Timestamp(stop)
        self.account_id = account_id
        self.portfolio_log_freq = portfolio_log_freq
        self.emitter = emitter
        self.strategy_state = strategy_state if strategy_state is not None else StrategyState()
        self.initializer = initializer
        self.warmup_mode = warmup_mode
        self._pregenerated_signals = dict()
        self._to_process = {}
        self._aux_data_reader = None

        # - get strategy parameters BEFORE simulation start
        #   potentially strategy may change it's parameters during simulation
        self.strategy_params = {}
        self.strategy_class = ""
        if self.setup.setup_type in [SetupTypes.STRATEGY, SetupTypes.STRATEGY_AND_TRACKER]:
            self.strategy_params = extract_parameters_from_object(self.setup.generator)
            self.strategy_class = full_qualified_class_name(self.setup.generator)

        self.ctx = self._create_backtest_context()

    def run(self, silent: bool = False, catch_keyboard_interrupt: bool = True, close_data_readers: bool = False):
        """
        Run the backtest from start to stop.

        Args:
            start (pd.Timestamp | str): The start time of the simulation.
            stop (pd.Timestamp | str): The end time of the simulation.
            silent (bool, optional): Whether to suppress progress output. Defaults to False.
        """
        logger.debug(f"[<y>SimulationRunner</y>] :: Running simulation from {self.start} to {self.stop}")

        self._prefetch_aux_data()

        # Start the context
        self.ctx.start()

        # Apply default warmup periods if strategy didn't set them
        for s in self.ctx.get_subscriptions():
            if not self.ctx.get_warmup(s) and (_d_wt := self.data_config.default_warmups.get(s)):
                logger.debug(
                    f"[<y>SimulationRunner</y>] :: Strategy didn't set warmup period for <c>{s}</c> so default <c>{_d_wt}</c> will be used"
                )
                self.ctx.set_warmup({s: _d_wt})

        # Subscribe to any custom data types if needed
        def _is_known_type(t: str):
            try:
                DataType(t)
                return True
            except:  # noqa: E722
                return False

        for t, r in self.data_config.data_providers.items():
            if not _is_known_type(t) or t in [
                DataType.TRADE,
                DataType.OHLC_TRADES,
                DataType.OHLC_QUOTES,
                DataType.QUOTE,
                DataType.ORDERBOOK,
            ]:
                logger.debug(f"[<y>BacktestContextRunner</y>] :: Subscribing to: {t}")
                self.ctx.subscribe(t, self.ctx.instruments)

        stop = self._stop or self.stop

        try:
            self._run(self.start, stop, silent=silent)
        except KeyboardInterrupt:
            logger.error("Simulated trading interrupted by user!")
            if not catch_keyboard_interrupt:
                raise
        except Exception as e:
            raise e
        finally:
            # Stop the context
            self.ctx.stop()
            if close_data_readers:
                for dp in self._data_providers:
                    for reader in dp._readers.values():
                        if hasattr(reader, "close"):
                            reader.close()  # type: ignore

    def _set_generated_signals(self, signals: pd.Series | pd.DataFrame):
        logger.debug(
            f"[<y>{self.__class__.__name__}</y>] :: Using pre-generated signals:\n {str(signals.count()).strip('ndtype: int64')}"
        )
        # - sanity check
        signals.index = pd.DatetimeIndex(signals.index)

        if isinstance(signals, pd.Series):
            self._pregenerated_signals[str(signals.name)] = signals  # type: ignore

        elif isinstance(signals, pd.DataFrame):
            for col in signals.columns:
                self._pregenerated_signals[col] = signals[col]  # type: ignore
        else:
            raise ValueError("Invalid signals or strategy configuration")

    def _prepare_generated_signals(self, start: str | pd.Timestamp, end: str | pd.Timestamp):
        for s, v in self._pregenerated_signals.items():
            _s_inst = None

            for i in self._data_providers[0].get_subscribed_instruments():
                # - we can process series with variable id's if we can find some similar instrument
                if s == i.symbol or s == str(i) or s == f"{i.exchange}:{i.symbol}" or str(s) == str(i):
                    _start, _end = np.datetime64(start), np.datetime64(end)
                    _start_idx, _end_idx = v.index.get_indexer([_start, _end], method="ffill")
                    sel = v.iloc[max(_start_idx, 0) : _end_idx + 1]

                    # TODO: check if data has exec_price - it means we have deals
                    self._to_process[i] = list(zip(sel.index, sel.values))
                    _s_inst = i
                    break

            if _s_inst is None:
                logger.error(f"Can't find instrument for pregenerated signals with id '{s}'")
                raise SimulationError(f"Can't find instrument for pregenerated signals with id '{s}'")

    def _process_generated_signals(self, instrument: Instrument, data_type: str, data: Any, is_hist: bool) -> bool:
        cc = self.channel
        t = np.datetime64(data.time, "ns")
        _account = self.account.get_account_processor(instrument.exchange)
        _data_provider = self._get_data_provider(instrument.exchange)
        assert isinstance(_account, SimulatedAccountProcessor)
        assert isinstance(_data_provider, SimulatedDataProvider)

        if not is_hist:
            # - signals for this instrument
            sigs = self._to_process[instrument]

            while sigs and t >= (_signal_time := sigs[0][0].as_unit("ns").asm8):
                self.time_provider.set_time(_signal_time)
                cc.send((instrument, "event", {"order": sigs[0][1]}, False))
                sigs.pop(0)

            if q := _account._exchange.emulate_quote_from_data(instrument, t, data):
                _data_provider._last_quotes[instrument] = q

        self.time_provider.set_time(t)
        cc.send((instrument, data_type, data, is_hist))

        return cc.control.is_set()

    def _process_strategy(self, instrument: Instrument, data_type: str, data: Any, is_hist: bool) -> bool:
        cc = self.channel
        t = np.datetime64(data.time, "ns")
        _account = self.account.get_account_processor(instrument.exchange)
        _data_provider = self._get_data_provider(instrument.exchange)
        assert isinstance(_account, SimulatedAccountProcessor)
        assert isinstance(_data_provider, SimulatedDataProvider)

        if not is_hist:
            if t >= (_next_exp_time := self.scheduler.next_expected_event_time()):
                # - we use exact event's time
                self.time_provider.set_time(_next_exp_time)
                self.scheduler.check_and_run_tasks()

            if q := _account._exchange.emulate_quote_from_data(instrument, t, data):
                _data_provider._last_quotes[instrument] = q

        self.time_provider.set_time(t)
        cc.send((instrument, data_type, data, is_hist))

        return cc.control.is_set()

    def _get_data_provider(self, exchange: str) -> IDataProvider:
        if exchange in self._exchange_to_data_provider:
            return self._exchange_to_data_provider[exchange]
        raise ValueError(f"Data provider for exchange {exchange} not found")

    def _run(self, start: pd.Timestamp, stop: pd.Timestamp, silent: bool = False) -> None:
        logger.info(f"{self.__class__.__name__} ::: Simulation started at {start} :::")

        if self._pregenerated_signals:
            self._prepare_generated_signals(start, stop)
            _run = self._process_generated_signals
        else:
            _run = self._process_strategy

        start, stop = pd.Timestamp(start), pd.Timestamp(stop)
        total_duration = stop - start
        update_delta = total_duration / 100
        prev_dt = np.datetime64(start)

        # - date iteration
        qiter = self._data_source.create_iterable(start, stop)

        if silent:
            for instrument, data_type, event, is_hist in qiter:
                # Handle NoDataContinue sentinel
                if isinstance(event, NoDataContinue):
                    if not self._handle_no_data_scenario(stop):
                        break
                    continue

                if not self._process_event(instrument, data_type, event, is_hist, _run, stop):
                    break
        else:
            _p = 0
            with tqdm(total=100, desc="Simulating", unit="%", leave=False) as pbar:
                for instrument, data_type, event, is_hist in qiter:
                    # Handle NoDataContinue sentinel
                    if isinstance(event, NoDataContinue):
                        if not self._handle_no_data_scenario(stop):
                            break
                        continue

                    if not self._process_event(instrument, data_type, event, is_hist, _run, stop):
                        break
                    dt = np.datetime64(event.time, "ns")
                    # update only if date has changed
                    if dt - prev_dt > update_delta:
                        _p += 1
                        pbar.n = _p
                        pbar.refresh()
                        prev_dt = dt
                pbar.n = 100
                pbar.refresh()

        logger.info(f"{self.__class__.__name__} ::: Simulation finished at {stop} :::")

    def _process_event(self, instrument, data_type, event, is_hist, _run, stop_time):
        """Process a single simulation event with proper time advancement and scheduler checks."""
        # During warmup, clamp future timestamps to current time
        if self.warmup_mode and hasattr(event, "time"):
            current_real_time = now_ns()
            if event.time > current_real_time:
                event.time = current_real_time

        if not _run(instrument, data_type, event, is_hist):
            return False

        return True

    def _handle_no_data_scenario(self, stop_time):
        """Handle scenario when no data is available but scheduler might have events."""
        # Check if we have pending scheduled events
        if hasattr(self.scheduler, "_next_nearest_time"):
            next_scheduled_time = self.scheduler._next_nearest_time
            current_time = self.time_provider.time()

            # Convert to int64 for numerical comparisons (avoid type issues)
            next_time_ns = next_scheduled_time.astype("int64")
            current_time_ns = current_time.astype("int64")
            stop_time_ns = stop_time.value  # Already int64

            # Check if we've reached the stop time
            if current_time_ns >= stop_time_ns:
                return False  # Stop simulation

            # If there's a scheduled event before stop time, advance to it
            if next_time_ns < np.iinfo(np.int64).max and next_time_ns < stop_time_ns:
                # Use the original datetime64 object for set_time (not the int64 conversion)
                self.time_provider.set_time(next_scheduled_time)
                self.scheduler.check_and_run_tasks()
                return True  # Continue simulation

        return False  # No scheduled events, stop simulation

    def print_latency_report(self) -> None:
        if (_l_r := SW.latency_report()) is not None:
            _llvl = QubxLogConfig.get_log_level()
            QubxLogConfig.set_log_level("INFO")
            logger.info(
                "<BLUE>   Time spent in simulation report   </BLUE>\n<r>"
                + _frame_to_str(
                    _l_r.sort_values("latency", ascending=False).reset_index(drop=True), "simulation", -1, -1, False
                )
                + "</r>"
            )
            QubxLogConfig.set_log_level(_llvl)

    def _create_backtest_context(self) -> IStrategyContext:
        logger.debug(
            f"[<y>Simulator</y>] :: Preparing simulated trading on <g>{self.setup.exchanges}</g> "
            f"for {self.setup.capital} {self.setup.base_currency}..."
        )

        channel = SimulatedCtrlChannel("databus", sentinel=(None, None, None, None))
        simulated_clock = SimulatedTimeProvider(np.datetime64(self.start, "ns"))

        account = self._construct_account_processor(
            self.setup.exchanges, self.setup.commissions, simulated_clock, channel
        )

        scheduler = SimulatedScheduler(channel, lambda: simulated_clock.time().item())

        data_source = IterableSimulationData(
            self.data_config.data_providers,
            open_close_time_indent_secs=self.data_config.adjusted_open_close_time_indent_secs,
        )

        brokers = []
        for exchange in self.setup.exchanges:
            _exchange_account = account.get_account_processor(exchange)
            assert isinstance(_exchange_account, SimulatedAccountProcessor)
            brokers.append(SimulatedBroker(channel, _exchange_account, _exchange_account._exchange))

        data_providers = []
        for exchange in self.setup.exchanges:
            _exchange_account = account.get_account_processor(exchange)
            assert isinstance(_exchange_account, SimulatedAccountProcessor)
            data_providers.append(
                SimulatedDataProvider(
                    exchange_id=exchange,
                    channel=channel,
                    scheduler=scheduler,
                    time_provider=simulated_clock,
                    account=_exchange_account,
                    readers=self.data_config.data_providers,
                    data_source=data_source,
                    open_close_time_indent_secs=self.data_config.adjusted_open_close_time_indent_secs,
                )
            )

        # - get aux data provider
        self._aux_data_reader = self.data_config.get_timeguarded_aux_reader(simulated_clock)

        # - it will store simulation results into memory
        logs_writer = InMemoryLogsWriter(self.account_id, self.setup.name, "0")

        # - it will store simulation results into memory
        strat: IStrategy | None = None

        match self.setup.setup_type:
            case SetupTypes.STRATEGY:
                strat = self.setup.generator  # type: ignore

            case SetupTypes.STRATEGY_AND_TRACKER:
                strat = self.setup.generator  # type: ignore
                strat.tracker = lambda ctx: self.setup.tracker  # type: ignore

            case SetupTypes.SIGNAL:
                strat = SignalsProxy(timeframe=self.setup.signal_timeframe)
                if len(data_providers) > 1:
                    raise SimulationConfigError("Signal setup is not supported for multiple exchanges !")

                self._set_generated_signals(self.setup.generator)  # type: ignore

                # - we don't need any unexpected triggerings
                self._stop = min(self.setup.generator.index[-1], self.stop)  # type: ignore

            case SetupTypes.SIGNAL_AND_TRACKER:
                strat = SignalsProxy(timeframe=self.setup.signal_timeframe)
                strat.tracker = lambda ctx: self.setup.tracker
                if len(data_providers) > 1:
                    raise SimulationConfigError("Signal setup is not supported for multiple exchanges !")

                self._set_generated_signals(self.setup.generator)  # type: ignore

                # - we don't need any unexpected triggerings
                self._stop = min(self.setup.generator.index[-1], self.stop)  # type: ignore

            case _:
                raise SimulationError(f"Unsupported setup type: {self.setup.setup_type} !")

        if not isinstance(strat, IStrategy):
            raise SimulationConfigError(f"Strategy should be an instance of IStrategy, but got {strat} !")

        ctx = StrategyContext(
            strategy=strat,
            brokers=brokers,
            data_providers=data_providers,
            account=account,
            scheduler=scheduler,
            time_provider=simulated_clock,
            instruments=self.setup.instruments,
            logging=StrategyLogging(logs_writer, portfolio_log_freq=self.portfolio_log_freq),
            aux_data_provider=self._aux_data_reader,
            emitter=self.emitter,
            strategy_state=self.strategy_state,
            initializer=self.initializer,
        )

        if self.emitter is not None:
            self.emitter.set_context(ctx)

        # - setup base subscription from spec
        if ctx.get_base_subscription() == DataType.NONE:
            logger.debug(
                f"[<y>simulator</y>] :: Setting up default base subscription: {self.data_config.default_base_subscription}"
            )
            ctx.set_base_subscription(self.data_config.default_base_subscription)

        # - set default on_event schedule if detected and strategy didn't set it's own schedule
        if not ctx.get_event_schedule("time") and self.data_config.default_trigger_schedule:
            logger.debug(f"[<y>simulator</y>] :: Setting default schedule: {self.data_config.default_trigger_schedule}")
            ctx.set_event_schedule(self.data_config.default_trigger_schedule)

        if self.setup.enable_funding:
            logger.debug("[<y>simulator</y>] :: Enabling funding rate simulation")
            ctx.subscribe(DataType.FUNDING_PAYMENT)

        self.logs_writer = logs_writer
        self.channel = channel
        self.time_provider = simulated_clock
        self.account = account
        self.scheduler = scheduler
        self._data_source = data_source
        self._data_providers = data_providers
        self._exchange_to_data_provider = {dp.exchange(): dp for dp in data_providers}
        return ctx

    def _construct_tcc(
        self, exchanges: list[str], commissions: str | dict[str, str | None] | None
    ) -> dict[str, TransactionCostsCalculator]:
        _exchange_to_tcc = {}
        if isinstance(commissions, (str, type(None))):
            commissions = {e: commissions for e in exchanges}
        for exchange in exchanges:
            _exchange_to_tcc[exchange] = lookup.find_fees(exchange.lower(), commissions.get(exchange))
        return _exchange_to_tcc

    def _construct_account_processor(
        self,
        exchanges: list[str],
        commissions: str | dict[str, str | None] | None,
        time_provider: ITimeProvider,
        channel: CtrlChannel,
    ) -> CompositeAccountProcessor:
        _exchange_to_tcc = self._construct_tcc(exchanges, commissions)
        for tcc in _exchange_to_tcc.values():
            if tcc is None:
                raise SimulationConfigError(
                    f"Can't find transaction costs calculator for '{self.setup.exchanges}' for specification '{self.setup.commissions}' !"
                )

        _exchange_to_simulated_exchange = {}
        for exchange in self.setup.exchanges:
            # - create simulated exchange:
            #   - we can use different emulations of real exchanges features in future here: for Binance, Bybit, InteractiveBrokers, etc.
            #   - for now we use simple basic simulated exchange implementation
            _exchange_to_simulated_exchange[exchange] = get_simulated_exchange(
                exchange, time_provider, _exchange_to_tcc[exchange], self.setup.accurate_stop_orders_execution
            )

        _account_processors = {}
        for exchange in self.setup.exchanges:
            _initial_capital = self.setup.capital
            if isinstance(_initial_capital, dict):
                _initial_capital = _initial_capital[exchange]
            assert isinstance(_initial_capital, (float, int))
            _account_processors[exchange] = SimulatedAccountProcessor(
                account_id=self.account_id,
                exchange=_exchange_to_simulated_exchange[exchange],
                channel=channel,
                base_currency=self.setup.base_currency,
                initial_capital=_initial_capital,
            )

        return CompositeAccountProcessor(
            time_provider=time_provider,
            account_processors=_account_processors,
        )

    def _prefetch_aux_data(self):
        # Perform prefetch of aux data if enabled
        if self._aux_data_reader is None:
            return

        aux_reader = self._aux_data_reader
        if isinstance(aux_reader, TimeGuardedWrapper) and isinstance(aux_reader._reader, CachedPrefetchReader):
            aux_reader = aux_reader._reader
        elif isinstance(aux_reader, CachedPrefetchReader):
            aux_reader = aux_reader
        else:
            return

        if self.data_config.prefetch_config and self.data_config.prefetch_config.enabled:
            # Prepare prefetch arguments
            prefetch_args = self.data_config.prefetch_config.args.copy()

            # Add exchange info if available from instruments
            if self.setup.instruments and "exchange" not in prefetch_args:
                # Get exchange from first instrument
                first_exchange = self.setup.instruments[0].exchange
                if first_exchange:
                    prefetch_args["exchange"] = first_exchange

            logger.info(
                f"Prefetching aux data: {self.data_config.prefetch_config.aux_data_names} for period {self.start} to {self.stop}"
            )

            try:
                # Perform the prefetch
                aux_reader.prefetch_aux_data(
                    self.data_config.prefetch_config.aux_data_names,
                    start=str(self.start),
                    stop=str(self.stop),
                    **prefetch_args,
                )
            except Exception as e:
                logger.warning(f"Prefetch failed: {e}")



"""Base template management system for strategy generation."""

import os
import shutil
from pathlib import Path
from typing import Any

import yaml
from jinja2 import Environment, FileSystemLoader, Template

from qubx import logger


class TemplateError(Exception):
    """Exception raised for template-related errors."""
    pass


class TemplateManager:
    """Manages strategy templates and generation."""
    
    def __init__(self):
        self.templates_dir = Path(__file__).parent
        self.built_in_templates = self._discover_built_in_templates()
    
    def _discover_built_in_templates(self) -> dict[str, Path]:
        """Discover all built-in templates."""
        templates = {}
        for template_dir in self.templates_dir.iterdir():
            if template_dir.is_dir() and not template_dir.name.startswith('_'):
                template_yml = template_dir / "template.yml"
                if template_yml.exists():
                    templates[template_dir.name] = template_dir
        return templates
    
    def list_templates(self) -> dict[str, dict]:
        """List all available built-in templates with their metadata."""
        templates_info = {}
        for name, template_path in self.built_in_templates.items():
            template_yml = template_path / "template.yml"
            if template_yml.exists():
                with open(template_yml) as f:
                    metadata = yaml.safe_load(f)
                templates_info[name] = metadata
            else:
                templates_info[name] = {"name": name, "description": "No description available"}
        return templates_info
    
    def generate_strategy(
        self,
        template_name: str | None = None,
        template_path: str | None = None,
        output_dir: str | Path = ".",
        **template_vars
    ) -> Path:
        """
        Generate a strategy from a template.
        
        Args:
            template_name: Name of built-in template to use
            template_path: Path to custom template directory
            output_dir: Directory to create strategy in
            **template_vars: Variables to substitute in templates
        
        Returns:
            Path to generated strategy directory
        """
        # Determine template directory
        if template_path:
            template_dir = Path(template_path)
            if not template_dir.exists():
                raise TemplateError(f"Template path does not exist: {template_path}")
        elif template_name:
            if template_name not in self.built_in_templates:
                available = list(self.built_in_templates.keys())
                raise TemplateError(f"Template '{template_name}' not found. Available: {available}")
            template_dir = self.built_in_templates[template_name]
        else:
            # Default to simple template
            template_name = "simple"
            if template_name not in self.built_in_templates:
                raise TemplateError("Default 'simple' template not found")
            template_dir = self.built_in_templates[template_name]
        
        # Load template metadata
        template_yml = template_dir / "template.yml"
        template_metadata = {}
        if template_yml.exists():
            with open(template_yml) as f:
                template_metadata = yaml.safe_load(f)
        
        # Set default template variables
        strategy_name = template_vars.get('name', 'my_strategy')
        default_vars = {
            'strategy_name': strategy_name,
            'strategy_class': self._to_class_name(strategy_name),
            'exchange': 'BINANCE.UM',
            'symbols': ['BTCUSDT'],
            'timeframe': '1h',
        }
        default_vars.update(template_vars)
        
        # Ensure symbols is a list
        if isinstance(default_vars['symbols'], str):
            default_vars['symbols'] = [s.strip() for s in default_vars['symbols'].split(',')]
        
        # Create output directory
        output_path = Path(output_dir).resolve() / strategy_name
        if output_path.exists():
            raise TemplateError(f"Directory already exists: {output_path}")
        
        logger.info(f"Creating strategy '{strategy_name}' from template '{template_dir.name}'")
        
        # Generate files
        self._render_template_directory(template_dir, output_path, default_vars)
        
        logger.info(f"Strategy created successfully at: {output_path}")
        return output_path
    
    def _render_template_directory(self, template_dir: Path, output_dir: Path, template_vars: dict):
        """Recursively render template directory to output directory."""
        # Create Jinja2 environment
        env = Environment(loader=FileSystemLoader(template_dir))
        
        # Walk through template directory
        for template_file in template_dir.rglob("*"):
            if template_file.is_file() and template_file.name != "template.yml":
                # Calculate relative path from template directory
                rel_path = template_file.relative_to(template_dir)
                
                # Create output file path with template variable substitution in path
                rel_path_str = str(rel_path)
                for var_name, var_value in template_vars.items():
                    rel_path_str = rel_path_str.replace(f"{{{{ {var_name} }}}}", str(var_value))
                
                output_file = output_dir / rel_path_str
                
                # Handle .j2 extension
                if output_file.suffix == '.j2':
                    output_file = output_file.with_suffix('')
                
                # Create parent directories
                output_file.parent.mkdir(parents=True, exist_ok=True)
                
                # Render template
                if template_file.suffix == '.j2':
                    # Render Jinja2 template
                    template = env.get_template(str(rel_path))
                    rendered_content = template.render(**template_vars)
                    with open(output_file, 'w') as f:
                        f.write(rendered_content)
                    
                    # Make shell scripts executable
                    if output_file.suffix == '.sh':
                        output_file.chmod(0o755)
                else:
                    # Copy non-template files as-is
                    shutil.copy2(template_file, output_file)
                
                logger.debug(f"Created: {output_file}")
    
    def _to_class_name(self, name: str) -> str:
        """Convert strategy name to proper class name."""
        # Convert snake_case or kebab-case to PascalCase
        words = name.replace('-', '_').split('_')
        return ''.join(word.capitalize() for word in words) + 'Strategy'


"""
Textual-based TUI for running qubx strategies with Jupyter kernel integration.

Features:
- Live REPL output panel for kernel interaction
- Application logs panel for qubx logger output
- Input bar for executing Python code in the kernel
- Real-time strategy monitoring
- Positions table with live updates
"""

from pathlib import Path

from qubx import logger

from .app import TextualStrategyApp

__all__ = ["run_strategy_yaml_in_textual"]


def run_strategy_yaml_in_textual(
    config_file: Path, account_file: Path | None = None, paper: bool = False, restore: bool = False
) -> None:
    """
    Run a strategy in a Textual TUI with Jupyter kernel integration.

    Args:
        config_file: Path to the strategy configuration file
        account_file: Path to the account configuration file
        paper: Whether to run in paper trading mode
        restore: Whether to restore the strategy state
    """
    if not config_file.exists():
        logger.error(f"Configuration file not found: {config_file}")
        return

    try:
        import nest_asyncio

        nest_asyncio.apply()
    except ImportError:
        logger.error("Can't find <r>nest_asyncio</r> module - try to install it first")
        return

    try:
        from jupyter_client import AsyncKernelManager  # noqa: F401
    except ImportError:
        logger.error("Can't find <r>jupyter_client</r> module - try to install it first")
        return

    logger.info("Running strategy in Textual TUI mode")

    # Create and run the app
    app = TextualStrategyApp(config_file, account_file, paper, restore)

    try:
        app.run()
    except Exception as e:
        logger.error(f"Textual app failed: {e}")
        raise



import re
from itertools import product
from types import FunctionType
from typing import Any, Callable, Type

import numpy as np

from qubx.utils.misc import generate_name


def _wrap_single_list(param_grid: list | dict) -> dict[str, Any] | list:
    """
    Wraps all non list values as single
    :param param_grid:
    :return:
    """
    as_list = lambda x: x if isinstance(x, (tuple, list, dict, np.ndarray)) else [x]  # noqa: E731
    if isinstance(param_grid, list):
        return [_wrap_single_list(ps) for ps in param_grid]
    return {k: as_list(v) for k, v in param_grid.items()}


def permutate_params(
    parameters: dict[str, list | tuple | Any],
    conditions: FunctionType | list | tuple | None = None,
    wrap_as_list=False,
) -> list[dict]:
    """
    Generate list of all permutations for given parameters and theirs possible values

    Example:

    >>> def foo(par1, par2):
    >>>     print(par1)
    >>>     print(par2)
    >>>
    >>> # permutate all values and call function for every permutation
    >>> [foo(**z) for z in permutate_params({
    >>>                                       'par1' : [1,2,3],
    >>>                                       'par2' : [True, False]
    >>>                                     }, conditions=lambda par1, par2: par1<=2 and par2==True)]

    1
    True
    2
    True

    :param conditions: list of filtering functions
    :param parameters: dictionary
    :param wrap_as_list: if True (default) it wraps all non list values as single lists (required for sklearn)
    :return: list of permutations
    """
    if conditions is None:
        conditions = []
    elif isinstance(conditions, FunctionType):
        conditions = [conditions]
    elif isinstance(conditions, (tuple, list)):
        if not all([isinstance(e, FunctionType) for e in conditions]):
            raise ValueError("every condition must be a function")
    else:
        raise ValueError("conditions must be of type of function, list or tuple")

    args = []
    vals = []
    for k, v in parameters.items():
        args.append(k)
        # vals.append([v] if not isinstance(v, (list, tuple)) else list(v) if isinstance(v, range) else v)
        match v:
            case list() | tuple():
                vals.append(v)
            case range():
                vals.append(list(v))
            case str():
                vals.append([v])
            case _:
                vals.append([v])
        # vals.append(v if isinstance(v, (List, Tuple)) else list(v) if isinstance(v, range) else [v])
    d = [dict(zip(args, p)) for p in product(*vals)]
    result = []
    for params_set in d:
        conditions_met = True
        for cond_func in conditions:
            func_param_args = cond_func.__code__.co_varnames
            func_param_values = [params_set[arg] for arg in func_param_args]
            if not cond_func(*func_param_values):
                conditions_met = False
                break
        if conditions_met:
            result.append(params_set)

    # if we need to follow sklearn rules we should wrap every noniterable as list
    return _wrap_single_list(result) if wrap_as_list else result


def dicts_product(d1: dict, d2: dict) -> dict:
    """
    Product of two dictionaries.

    Example:
    -------

    dicts_product({
        'A': 1,
        'B': 2,
    }, {
        'C': 3,
        'D': 4,
    })

    Output:
    ------
    {
        'A + C': [1, 3],
        'A + D': [1, 4],
        'B + C': [2, 3],
        'B + D': [2, 4]
    }

    """

    def flatten(lst):
        return [item for sublist in lst for item in (sublist if isinstance(sublist, list) else [sublist])]

    return {(a + " + " + b): flatten([d1[a], d2[b]]) for a, b in product(d1.keys(), d2.keys())}


class _dict(dict):
    def __add__(self, other: dict) -> dict:
        return _dict(dicts_product(self, other))


def variate(clz: Type[Any] | list[Type[Any]], *args, conditions=None, **kwargs) -> _dict:
    """
    Make variations of parameters for simulations (micro optimizer)

    Example:

    >>>    class MomentumStrategy_Ex1_test:
    >>>       def __init__(self, p1, lookback_period=10, filter_type='sma', skip_entries_flag=False):
    >>>            self.p1, self.lookback_period, self.filter_type, self.skip_entries_flag = p1, lookback_period, filter_type, skip_entries_flag
    >>>
    >>>        def __repr__(self):
    >>>            return self.__class__.__name__ + f"({self.p1},{self.lookback_period},{self.filter_type},{self.skip_entries_flag})"
    >>>
    >>>    variate(MomentumStrategy_Ex1_test, 10, lookback_period=[1,2,3], filter_type=['ema', 'sma'], skip_entries_flag=[True, False])

    Output:
    >>>    {
    >>>        'MSE1t_(lp=1,ft=ema,sef=True)':  MomentumStrategy_Ex1_test(10,1,ema,True),
    >>>        'MSE1t_(lp=1,ft=ema,sef=False)': MomentumStrategy_Ex1_test(10,1,ema,False),
    >>>        'MSE1t_(lp=1,ft=sma,sef=True)':  MomentumStrategy_Ex1_test(10,1,sma,True),
    >>>        'MSE1t_(lp=1,ft=sma,sef=False)': MomentumStrategy_Ex1_test(10,1,sma,False),
    >>>        'MSE1t_(lp=2,ft=ema,sef=True)':  MomentumStrategy_Ex1_test(10,2,ema,True),
    >>>        'MSE1t_(lp=2,ft=ema,sef=False)': MomentumStrategy_Ex1_test(10,2,ema,False),
    >>>        'MSE1t_(lp=2,ft=sma,sef=True)':  MomentumStrategy_Ex1_test(10,2,sma,True),
    >>>        'MSE1t_(lp=2,ft=sma,sef=False)': MomentumStrategy_Ex1_test(10,2,sma,False),
    >>>        'MSE1t_(lp=3,ft=ema,sef=True)':  MomentumStrategy_Ex1_test(10,3,ema,True),
    >>>        'MSE1t_(lp=3,ft=ema,sef=False)': MomentumStrategy_Ex1_test(10,3,ema,False),
    >>>        'MSE1t_(lp=3,ft=sma,sef=True)':  MomentumStrategy_Ex1_test(10,3,sma,True),
    >>>        'MSE1t_(lp=3,ft=sma,sef=False)': MomentumStrategy_Ex1_test(10,3,sma,False)
    >>>    }

    and using in simuation:

    >>>    r = simulate(
    >>>             variate(MomentumStrategy_Ex1_test, 10, lookback_period=[1,2,3], filter_type=['ema', 'sma'], skip_entries_flag=[True, False]),
    >>>             data, capital, ["BINANCE.UM:BTCUSDT"], dict(type="ohlc", timeframe="5Min", nback=0), "5Min -1Sec", "vip0_usdt", "2024-01-01", "2024-01-02"
    >>>    )

    Also it's possible to pass a class with tracker:
    >>>    variate([MomentumStrategy_Ex1_test, AtrTracker(2, 1)], 10, lookback_period=[1,2,3], filter_type=['ema', 'sma'], skip_entries_flag=[True, False])
    """

    def _cmprss(xs: str):
        return "".join([x[0] for x in re.split(r"((?<!-)(?=[A-Z]))|_|(\d)", xs) if x])

    if isinstance(clz, (type, Callable)):
        sfx = _cmprss(clz.__name__)
        _mk = lambda k, *args, **kwargs: k(*args, **kwargs)  # noqa: E731
    elif isinstance(clz, (list, tuple)) and clz and isinstance(clz[0], type):
        sfx = _cmprss(clz[0].__name__)
        _mk = lambda k, *args, **kwargs: [k[0](*args, **kwargs), *k[1:]]  # noqa: E731
    else:
        raise ValueError(
            "Can't recognize data for variating: must be either a class type or a list where first element is class type"
        )

    def _v_to_str(x: Any) -> str:
        if isinstance(x, (list, tuple, dict, set, np.ndarray)) and len(xs := str(x)) > 15:
            return "[" + generate_name(xs, 8).lower() + "]"
        return str(x)

    to_excl = [s for s, v in kwargs.items() if not isinstance(v, (list, set, tuple, range))]
    dic2str = lambda ds: [_cmprss(k) + "=" + _v_to_str(v) for k, v in ds.items() if k not in to_excl]  # noqa: E731

    return _dict(
        {
            f"{sfx}_({','.join(dic2str(z))})": _mk(clz, *args, **z)
            for z in permutate_params(kwargs, conditions=conditions)
        }
    )



from typing import Any, Optional

import numpy as np

from qubx.core.basics import TargetPosition, dt_64
from qubx.core.interfaces import IAccountViewer
from qubx.exporters.formatters.base import DefaultFormatter


class TargetPositionFormatter(DefaultFormatter):
    """
    Formatter for exporting target positions as structured messages.

    This formatter creates messages suitable for trading systems that need
    to know target position sizes with leverage calculations.
    """

    def __init__(
        self,
        alert_name: str,
        exchange_mapping: Optional[dict[str, str]] = None,
        account: Optional[IAccountViewer] = None,
    ):
        """
        Initialize the TargetPositionFormatter.

        Args:
            alert_name: The name of the alert to include in the messages
            exchange_mapping: Optional mapping of exchange names to use in messages.
                             If an exchange is not in the mapping, the instrument's exchange is used.
            account: The account viewer to get account information like total capital, leverage, etc.
        """
        super().__init__()
        self.alert_name = alert_name
        self.exchange_mapping = exchange_mapping or {}

    def format_target_position(self, time: dt_64, target: TargetPosition, account: IAccountViewer) -> dict[str, Any]:
        """
        Format a target position for export.

        This method creates a structured message with target position information
        including leverage calculated as (notional size / total capital).

        Args:
            time: Timestamp when the target position was generated
            target: The target position to format
            account: Account viewer to get account information like total capital, leverage, etc.

        Returns:
            A dictionary containing the formatted target position data
        """
        # Get price: use entry_price if available, else fallback to position's last_update_price
        price = target.entry_price
        if price is None:
            position = account.get_position(target.instrument)
            if not np.isnan(position.last_update_price):
                price = position.last_update_price
            else:
                # Cannot calculate leverage without price
                return {}

        # Calculate notional size and leverage
        notional = abs(target.target_position_size * price)
        total_capital = account.get_total_capital(exchange=target.instrument.exchange)
        leverage = notional / total_capital if total_capital > 0 else 0.0

        # Determine side
        side = "BUY" if target.target_position_size > 0 else "SELL"

        # Get the exchange name from mapping or use the instrument's exchange
        exchange = self.exchange_mapping.get(target.instrument.exchange, target.instrument.exchange)

        return {
            "type": "TARGET_POSITION",
            "data": f'{{"action":"TARGET_POSITION","alertName":"{self.alert_name}","exchange":"{exchange}","symbol":"{target.instrument.exchange_symbol.upper()}","side":"{side}","leverage":{leverage}}}',
        }



"""
Strategy initialization code generation for Jupyter kernel.

Generates Python code to initialize strategy context and helper functions
that will be executed in the kernel.
"""

from pathlib import Path


def generate_init_code(
    config_file: Path,
    account_file: Path | None,
    paper: bool,
    restore: bool,
) -> str:
    """
    Generate initialization code to inject strategy context into the kernel.

    Args:
        config_file: Path to the strategy configuration file
        account_file: Optional path to the account configuration file
        paper: Whether to run in paper trading mode
        restore: Whether to restore the strategy state

    Returns:
        Python code string to be executed in the kernel
    """
    # Convert paths to strings for the code generation
    config_path_str = str(config_file.absolute())
    account_path_str = str(account_file.absolute()) if account_file else "None"

    return f"""
import pandas as pd
import sys
from pathlib import Path
from qubx import logger
from qubx.core.basics import Instrument, Position
from qubx.core.context import StrategyContext
from qubx.core.interfaces import IStrategyContext
from qubx.utils.misc import dequotify, add_project_to_system_path
from qubx.utils.runner.runner import run_strategy_yaml
import nest_asyncio

# Apply nest_asyncio for nested event loops
nest_asyncio.apply()

pd.set_option('display.max_colwidth', None, 'display.max_columns', None, 'display.width', 1000)

# Initialize the strategy context
config_file = Path('{config_path_str}')
account_file = Path('{account_path_str}') if '{account_path_str}' != 'None' else None

# Add project to system path
add_project_to_system_path()
add_project_to_system_path(str(config_file.parent.parent))
add_project_to_system_path(str(config_file.parent))

# Run the strategy
ctx = run_strategy_yaml(config_file, account_file, paper={paper}, restore={restore}, blocking=False)
S = ctx.strategy

def _pos_to_dict(p: Position):
    mv = round(p.notional_value, 3)
    return dict(
        MktValue=mv,
        Position=round(p.quantity, p.instrument.size_precision),
        PnL=p.total_pnl(),
        AvgPrice=round(p.position_avg_price_funds, p.instrument.price_precision),
        LastPrice=round(p.last_update_price, p.instrument.price_precision),
    )

def portfolio(all=True):
    from tabulate import tabulate
    d = dict()
    for s, p in ctx.get_positions().items():
        if p.quantity != 0.0 or all:
            d[dequotify(s.symbol, s.quote)] = _pos_to_dict(p)
    d = pd.DataFrame.from_dict(d, orient='index')
    if d.empty:
        print('-(no open positions yet)-')
        return
    d = d.sort_values('MktValue', ascending=False)
    print(tabulate(d, ['MktValue', 'Position', 'PnL', 'AvgPrice', 'LastPrice'], tablefmt='rounded_grid'))

def orders():
    if (_orders := ctx.get_orders()):
        print("- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -")
        for k, (i, o) in enumerate(_orders.items()):
            print(f" [{{k}}] {{i}} {{o.status}} {{o.side}} {{o.instrument.symbol}} {{o.quantity}} @ {{o.price}} - {{o.time}}")
        print("- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -")

def trade(instrument, qty: float, price=None, tif='gtc'):
    return ctx.trade(instrument if isinstance(instrument, Instrument) else instrument._instrument, qty, price, tif)

__exit = exit
def exit():
    ctx.stop()
    __exit()

# Positions emit helper for Textual UI
from IPython.display import display

_CUSTOM_MIME_POS = "application/x-qubx-positions+json"

def _positions_as_records(all=True):
    rows = []
    try:
        for s, p in ctx.get_positions().items():
            if p.quantity != 0.0 or all:
                # Handle NaN values by converting to None or 0
                pnl = p.total_pnl()
                if not isinstance(pnl, (int, float)) or (isinstance(pnl, float) and (pnl != pnl or pnl == float('inf') or pnl == float('-inf'))):
                    pnl = 0.0

                mkt_value = p.notional_value
                if not isinstance(mkt_value, (int, float)) or (isinstance(mkt_value, float) and (mkt_value != mkt_value or mkt_value == float('inf') or mkt_value == float('-inf'))):
                    mkt_value = 0.0

                rows.append({{
                    "symbol": s.symbol,
                    "side": "LONG" if p.quantity > 0 else ("SHORT" if p.quantity < 0 else "FLAT"),
                    "qty": round(p.quantity, s.size_precision),
                    "avg_px": round(p.position_avg_price_funds, s.price_precision),
                    "last_px": round(p.last_update_price, s.price_precision),
                    "pnl": round(pnl, 2),
                    "mkt_value": round(mkt_value, 3),
                }})
    except Exception:
        pass  # Context not ready yet
    return rows

def emit_positions(all=True):
    \"\"\"Publish current positions via custom MIME for Textual to capture.\"\"\"
    try:
        data = {{ _CUSTOM_MIME_POS: _positions_as_records(all=all) }}
        display(data, raw=True)
    except Exception:
        pass  # Silently fail if context is not ready

print(f"Strategy initialized: {{ctx.strategy.__class__.__name__}}")
print(f"Instruments: {{[i.symbol for i in ctx.instruments]}}")
"""



from typing import Any

import pandas as pd

from qubx import logger
from qubx.connectors.ccxt.broker import CcxtBroker
from qubx.connectors.ccxt.utils import ccxt_convert_order_info, instrument_to_ccxt_symbol
from qubx.core.basics import (
    Instrument,
    Order,
    OrderSide,
)
from qubx.core.exceptions import BadRequest


class HyperliquidCcxtBroker(CcxtBroker):
    """
    HyperLiquid-specific broker that handles market order slippage requirements.

    HyperLiquid requires a price even for market orders to calculate max slippage.
    This broker automatically calculates slippage-protected prices for market orders.
    """

    def __init__(
        self,
        *args,
        market_order_slippage: float = 0.05,  # 5% default slippage
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.market_order_slippage = market_order_slippage

    def _enrich_order_response(
        self,
        response: dict[str, Any],
        symbol: str,
        order_type: str,
        side: str,
        amount: float,
        price: float | None = None,
        client_id: str | None = None,
        params: dict[str, Any] | None = None,
    ) -> dict[str, Any]:
        """
        Fill in missing fields in HyperLiquid order response.

        HyperLiquid often returns minimal order information (sometimes only order ID),
        so we need to reconstruct the full order from the request parameters.

        Args:
            response: The raw response from HyperLiquid
            symbol: CCXT symbol (e.g., "BTC/USDC:USDC")
            order_type: Order type ("limit" or "market")
            side: Order side ("buy" or "sell")
            amount: Order amount
            price: Order price (optional for market orders)
            client_id: Client order ID (optional)
            params: Additional params (timeInForce, reduceOnly, etc.)

        Returns:
            Enriched response dictionary ready for ccxt_convert_order_info
        """
        if params is None:
            params = {}

        # Fill in missing fields from the request parameters
        if not response.get("symbol"):
            response["symbol"] = symbol
        if not response.get("type"):
            response["type"] = order_type
        # Always set side from request params (CCXT can't parse from resting/filled structures)
        if not response.get("side") or response.get("side") == "unknown":
            response["side"] = side
        if not response.get("amount") or response.get("amount") == 0:
            response["amount"] = amount
        if not response.get("price") or response.get("price") == 0:
            if price is not None:
                response["price"] = price
        if not response.get("timestamp"):
            response["timestamp"] = pd.Timestamp.now(tz="UTC").value // 1_000_000  # Convert to milliseconds
        if not response.get("status"):
            response["status"] = "open"
        if not response.get("timeInForce"):
            response["timeInForce"] = params.get("timeInForce")
        if not response.get("clientOrderId") and client_id:
            response["clientOrderId"] = client_id
        if not response.get("cost"):
            response["cost"] = 0.0

        # Ensure reduceOnly is propagated
        if params.get("reduceOnly"):
            response["reduceOnly"] = True

        return response

    def _translate_time_in_force(self, tif: str | None) -> str | None:
        """
        Translate standard timeInForce values to HyperLiquid format.

        HyperLiquid uses:
        - "Alo" for Add Liquidity Only (post-only/maker)
        - "Ioc" for Immediate or Cancel
        - "Gtc" for Good til Cancelled

        This method translates common aliases:
        - GTX (Binance post-only) → Alo
        - FOK (Fill or Kill) → Ioc
        - GTC, IOC, ALO (any case) → proper HyperLiquid format
        """
        if tif is None:
            return None

        # If already in correct HyperLiquid format, return as-is
        if tif in ["Alo", "Ioc", "Gtc"]:
            return tif

        # Translation map for timeInForce values
        tif_map = {
            "GTX": "Alo",  # Binance post-only → Hyperliquid post-only
            "FOK": "Ioc",  # Fill or Kill → Immediate or Cancel
            "GTC": "Gtc",  # Good til Cancelled (normalize case)
            "IOC": "Ioc",  # Immediate or Cancel (normalize case)
            "ALO": "Alo",  # Add Liquidity Only (normalize case)
        }

        # Normalize to uppercase for lookup
        tif_upper = tif.upper()
        return tif_map.get(tif_upper, tif)  # Return translated or original if not in map

    async def _create_order(
        self,
        instrument: Instrument,
        order_side: OrderSide,
        order_type: str,
        amount: float,
        price: float | None = None,
        client_id: str | None = None,
        time_in_force: str = "gtc",
        **options,
    ) -> tuple[Order | None, Exception | None]:
        """
        Override _create_order to fill missing order details from request.

        HyperLiquid returns minimal order information (only order ID),
        so we need to reconstruct the full order from the request parameters.
        """
        try:
            payload = self._prepare_order_payload(
                instrument, order_side, order_type, amount, price, client_id, time_in_force, **options
            )
            if self.enable_create_order_ws:
                r = await self._exchange_manager.exchange.create_order_ws(**payload)
            else:
                r = await self._exchange_manager.exchange.create_order(**payload)

            if r is None:
                msg = "(::_create_order) No response from exchange"
                logger.error(msg)
                from ccxt.base.errors import ExchangeError

                return None, ExchangeError(msg)

            if r["id"] is None:
                return None, None

            # Fill in missing fields from the request parameters
            r = self._enrich_order_response(
                response=r,
                symbol=payload["symbol"],
                order_type=payload["type"],
                side=payload["side"],
                amount=payload["amount"],
                price=payload["price"],
                client_id=client_id,
                params=payload["params"],
            )

            order = ccxt_convert_order_info(instrument, r)
            logger.info(f"New order {order}")
            return order, None

        except Exception as err:
            return None, err

    def _prepare_order_payload(
        self,
        instrument: Instrument,
        order_side: OrderSide,
        order_type: str,
        amount: float,
        price: float | None = None,
        client_id: str | None = None,
        time_in_force: str = "gtc",
        **options,
    ) -> dict[str, Any]:
        # Handle market orders that need slippage-protected prices
        if order_type.lower() == "market" and price is None:
            quote = self.data_provider.get_quote(instrument)
            if quote is None:
                logger.warning(
                    f"[<y>{instrument.symbol}</y>] :: Quote is not available for market order slippage calculation."
                )
                raise BadRequest(
                    f"Quote is not available for market order slippage calculation for {instrument.symbol}"
                )

            # Get slippage from options or use default
            slippage = options.get("slippage", self.market_order_slippage)

            # Calculate slippage-protected price
            if order_side.upper() == "BUY":
                # For buy orders, add slippage to ask price to ensure execution
                price = quote.ask * (1 + slippage)
                logger.debug(
                    f"[<y>{instrument.symbol}</y>] :: Market BUY order: using slippage-protected price {price:.6f} (ask: {quote.ask:.6f}, slippage: {slippage:.1%})"
                )
            else:  # SELL
                # For sell orders, subtract slippage from bid price to ensure execution
                price = quote.bid * (1 - slippage)
                logger.debug(
                    f"[<y>{instrument.symbol}</y>] :: Market SELL order: using slippage-protected price {price:.6f} (bid: {quote.bid:.6f}, slippage: {slippage:.1%})"
                )

        # Call parent implementation with calculated price
        payload = super()._prepare_order_payload(
            instrument, order_side, order_type, amount, price, client_id, time_in_force, **options
        )

        # Translate timeInForce to HyperLiquid format (GTX → Alo, etc.)
        params = payload.get("params", {})
        if "timeInForce" in params:
            params["timeInForce"] = self._translate_time_in_force(params["timeInForce"])

        # Add slippage parameter to params if specified in options
        if "slippage" in options:
            # HyperLiquid accepts slippage as a percentage (e.g., 0.05 for 5%)
            params["px"] = price  # Explicit price for slippage calculation

        payload["params"] = params
        return payload

    async def _edit_order_async(
        self, order_id: str, existing_order: Order, price: float, amount: float
    ) -> tuple[Order | None, Exception | None]:
        """
        Override _edit_order_async with WebSocket support and response enrichment.

        HyperLiquid requires params to match the original order and may return
        minimal order information, so we fill missing fields from the request.
        """
        try:
            ccxt_symbol = instrument_to_ccxt_symbol(existing_order.instrument)
            ccxt_side = "buy" if existing_order.side == "BUY" else "sell"

            params = {}
            if existing_order.time_in_force:
                # Translate TIF to HyperLiquid format (GTX → Alo, etc.)
                params["timeInForce"] = self._translate_time_in_force(existing_order.time_in_force)
            else:
                # If no TIF is set on the existing order, default to Alo (post-only) for limit orders
                # This is the safest default for market making
                params["timeInForce"] = "Alo"
            if existing_order.options.get("reduceOnly", False):
                params["reduceOnly"] = True

            # Use WebSocket if enabled, otherwise use REST API
            if self.enable_edit_order_ws:
                result = await self._exchange_manager.exchange.edit_order_ws(
                    id=order_id,
                    symbol=ccxt_symbol,
                    type="limit",
                    side=ccxt_side,
                    amount=amount,
                    price=price,
                    params=params,
                )
            else:
                result = await self._exchange_manager.exchange.edit_order(
                    id=order_id,
                    symbol=ccxt_symbol,
                    type="limit",
                    side=ccxt_side,
                    amount=amount,
                    price=price,
                    params=params,
                )

            # Fill in missing fields from the request parameters
            result = self._enrich_order_response(
                response=result,
                symbol=ccxt_symbol,
                order_type="limit",
                side=ccxt_side,
                amount=amount,
                price=price,
                client_id=existing_order.client_id,
                params=params,
            )

            # Convert the result back to our Order format
            updated_order = ccxt_convert_order_info(existing_order.instrument, result)
            return updated_order, None

        except Exception as err:
            logger.error(f"Async edit order failed for {order_id}: {err}")
            return None, err



import asyncio
import json
import threading
from os.path import exists, expanduser
from pathlib import Path
from queue import Queue
from threading import Thread
from typing import Any, Iterable

import aiohttp
import numpy as np
import orjson
import pandas as pd
from pyarrow import csv

from qubx import logger
from qubx.data.registry import reader
from qubx.utils.misc import AsyncThreadLoop
from qubx.utils.ntp import time_now
from qubx.utils.orderbook import accumulate_orderbook_levels
from qubx.utils.time import handle_start_stop

from .readers import DataReader, DataTransformer

TARDIS_EXCHANGE_MAPPERS = {
    "bitfinex.f": "bitfinex-derivatives",
    "binance.um": "binance-futures",
    "binance.pm": "binance-futures",
}


class TardisCsvDataReader(DataReader):
    def __init__(self, path: str | Path) -> None:
        _path = expanduser(path)
        if not exists(_path):
            raise ValueError(f"Folder is not found at {path}")
        self.path = Path(_path)

    def get_names(self, exchange: str | None = None, data_type: str | None = None) -> list[str]:
        symbols = []
        exchanges = [exchange] if exchange else self.get_exchanges()
        for exchange in exchanges:
            exchange_path = Path(self.path) / exchange
            if not exists(exchange_path):
                raise ValueError(f"Exchange is not found at {exchange_path}")
            data_types = [data_type] if data_type else self.get_data_types(exchange)
            for data_type in data_types:
                data_type_path = exchange_path / data_type
                if not exists(data_type_path):
                    return []
                symbols += self._get_symbols(data_type_path)
        return symbols

    def read(
        self,
        data_id: str,
        start: str | None = None,
        stop: str | None = None,
        transform: DataTransformer = DataTransformer(),
        chunksize=0,
        timeframe=None,
        data_type="trades",
    ) -> Iterable | Any:
        if chunksize > 0:
            raise NotImplementedError("Chunksize is not supported for TardisCsvDataReader")
        exchange, symbol = data_id.split(":")
        _exchange = exchange.lower()
        _exchange = TARDIS_EXCHANGE_MAPPERS.get(_exchange, _exchange)
        t_0, t_1 = handle_start_stop(start, stop, lambda x: pd.Timestamp(x).date().isoformat())
        _path = self.path / _exchange / data_type
        if not _path.exists():
            raise ValueError(f"Data type is not found at {_path}")
        _files = sorted(_path.glob(f"*_{symbol}.csv.gz"))
        if not _files:
            return None
        _dates = [file.stem.split("_")[0] for file in _files]
        if t_0 is None:
            t_0 = _dates[0]
        if t_1 is None:
            t_1 = _dates[-1]
        _filt_files = [file for file in _files if t_0 <= file.stem.split("_")[0] <= t_1]

        tables = []
        # fieldnames = None
        for f_path in _filt_files:
            table = pd.read_csv(f_path)
            tables.append(table)
            # table = csv.read_csv(
            #     f_path,
            #     parse_options=csv.ParseOptions(ignore_empty_lines=True),
            # )
            # if not fieldnames:
            #     fieldnames = table.column_names
            # tables.append(table.to_pandas())

        return pd.concat(tables)

        # transform.start_transform(data_id, fieldnames or [], start=start, stop=stop)
        # raw_data = pd.concat(tables).to_numpy()
        # transform.process_data(raw_data)

        # return transform.collect()

    def get_exchanges(self) -> list[str]:
        return [exchange.name for exchange in self.path.iterdir() if exchange.is_dir()]

    def get_data_types(self, exchange: str) -> list[str]:
        exchange_path = Path(self.path) / exchange
        return [data_type.name for data_type in exchange_path.iterdir() if data_type.is_dir()]

    def _get_symbols(self, data_type_path: Path) -> list[str]:
        symbols = set()
        for file in data_type_path.glob("*.gz"):
            parts = file.stem.replace(".csv", "").split("_")
            if len(parts) == 2:
                symbols.add(parts[1])
        return list(symbols)


@reader("tardis")
class TardisMachineReader(DataReader):
    """
    Data reader for Tardis.dev API which provides access to historical cryptocurrency market data.

    This reader connects to both the Tardis.dev HTTP API for exchange information and
    the Tardis Machine server for data retrieval using the normalized API endpoint.

    Parameters:
    -----------
    api_key : str | None
        Optional API key for authentication with Tardis.dev API. If not provided,
        only the first day of each month of data is accessible (free access).
    tardis_api_url : str
        URL for the Tardis.dev API (default: "https://api.tardis.dev/v1")
    tardis_machine_url : str
        URL for the Tardis Machine server (default: "http://quantlab:8010")
    """

    # Define the standard column names for each normalized data type
    NORMALIZED_DATA_COLUMNS = {
        "trade": ["timestamp", "type", "symbol", "exchange", "side", "price", "amount", "localTimestamp"],
        "book_change": ["timestamp", "type", "symbol", "exchange", "side", "price", "amount", "localTimestamp"],
        "book_snapshot": [
            "timestamp",
            "top_bid",
            "top_ask",
            "tick_size",
            "bids",
            "asks",
        ],
        "derivative_ticker": [
            "timestamp",
            "type",
            "symbol",
            "exchange",
            "fundingRate",
            "fundingTimestamp",
            "openInterest",
            "localTimestamp",
        ],
        "trade_bar": [
            "timestamp",
            "type",
            "symbol",
            "exchange",
            "name",
            "interval",
            "kind",
            "open",
            "high",
            "low",
            "close",
            "volume",
            "buyVolume",
            "sellVolume",
            "trades",
            "vwap",
            "openTimestamp",
            "closeTimestamp",
            "localTimestamp",
        ],
        "disconnect": ["type", "exchange", "localTimestamp"],
    }

    def __init__(
        self,
        tardis_api_url: str = "https://api.tardis.dev/v1",
        tardis_machine_url: str = "http://quantlab:8010",
    ) -> None:
        self.api_url = tardis_api_url.rstrip("/")
        self.machine_url = tardis_machine_url.rstrip("/")
        self._exchanges_cache = None
        self._exchange_info_cache = {}

        # Create asyncio loop and start it in a thread
        self._loop = asyncio.new_event_loop()
        self._thread = Thread(target=self._start_loop, daemon=True)
        self._thread.start()
        self._async_loop = AsyncThreadLoop(self._loop)

    def __del__(self):
        """Cleanup resources"""
        if self._loop and not self._loop.is_closed():
            self._loop.call_soon_threadsafe(self._loop.stop)
            self._thread.join(timeout=5)
            self._loop.close()

    async def _fetch_stream(self, url: str, line_queue: Queue, stop_event: threading.Event):
        """Fetch streaming data using aiohttp"""
        # Create a new session for this read operation
        timeout = aiohttp.ClientTimeout(total=0)
        session = aiohttp.ClientSession(headers=self._get_headers(), timeout=timeout)
        try:
            async with session.get(url) as response:
                if response.status != 200:
                    error_text = await response.text()
                    logger.error(f"Error fetching data: {error_text}")
                    return

                async for line in response.content:
                    if stop_event.is_set():
                        break
                    if not line.strip():
                        continue
                    line_queue.put(line.decode())
        except Exception as e:
            logger.error(f"Error in fetch stream: {e} url: {url}", exc_info=True)
        finally:
            line_queue.put(None)  # Signal end of stream
            await session.close()  # Ensure session is closed

    async def _http_get(self, url: str) -> tuple[int, dict | str]:
        """Make an HTTP GET request using aiohttp"""
        # Create a new session for this request
        session = aiohttp.ClientSession(headers=self._get_headers())
        try:
            async with session.get(url) as response:
                if response.status == 200:
                    return response.status, await response.json()
                else:
                    return response.status, await response.text()
        except Exception as e:
            logger.error(f"HTTP request failed: {e}")
            return 500, str(e)
        finally:
            await session.close()  # Ensure session is closed

    def _start_loop(self):
        """Start the asyncio event loop in the thread"""
        asyncio.set_event_loop(self._loop)
        self._loop.run_forever()

    def _get_headers(self) -> dict:
        """Get headers for API requests including authentication if available"""
        headers = {"Content-Type": "application/json"}
        return headers

    def _safe_log_response(self, response):
        """Safely log response content without HTML tags that could cause issues with logger"""
        try:
            # Try to parse as JSON first
            return str(response.json())
        except:
            # If not JSON, return status code only to avoid HTML parsing issues
            return f"Status code: {response.status_code}"

    def _get_normalized_data_type(self, data_type: str, timeframe: str | None = None) -> list[str]:
        """
        Map input data type to normalized data types used by Tardis Machine API

        Parameters:
        -----------
        data_type : str
            Input data type (e.g., 'trade', 'orderBookL2', 'book')
        timeframe : str | None
            Optional timeframe for aggregation (e.g., '1m', '5m')

        Returns:
        --------
        list[str]
            List of normalized data types to request
        """
        if timeframe:
            return [f"trade_bar_{timeframe}"]

        # Map common data types to normalized data types
        if data_type in ["orderBookL2", "book", "orderbook"]:
            # For now, we'll only support book_snapshot and not book_change
            # as book_change can have isSnapshot=True which requires special handling
            # TODO: maybe change to 1000 levels and 1000ms interval
            return ["book_snapshot_500_1000ms"]
        elif data_type == "book_change":
            raise NotImplementedError(
                "book_change data type is not implemented yet due to complexity with isSnapshot handling"
            )
        elif data_type in ["trade", "trades"]:
            return ["trade"]
        elif data_type == "derivative_ticker":
            return ["derivative_ticker"]
        elif data_type.startswith("book_snapshot_"):
            # Allow direct specification of book_snapshot with levels and interval
            return [data_type]
        else:
            # Default to the provided data type
            return [data_type]

    def _get_column_names(self, data_type: str) -> list[str]:
        """
        Get standard column names for a normalized data type

        Parameters:
        -----------
        data_type : str
            Normalized data type (e.g., 'trade', 'book_change')

        Returns:
        --------
        list[str]
            List of column names in the standard order
        """
        # Extract the base type from data types like 'trade_bar_1m'
        base_type = data_type
        if data_type.startswith("trade_bar_"):
            base_type = "trade_bar"
        elif data_type.startswith(("book_snapshot_", "orderbook")):
            base_type = "book_snapshot"

        return self.NORMALIZED_DATA_COLUMNS.get(base_type, [])

    def _parse_record(self, record: dict, data_type: str, tick_size_pct: float = 0.01, depth: int = 100) -> list | None:
        """
        Parse a record from the normalized API response into a list of values

        Parameters:
        -----------
        record : dict
            Record from the normalized API response
        data_type : str
            Normalized data type of the record

        Returns:
        --------
        list
            List of values in the standard order for the data type
        """
        if data_type == "book_snapshot":
            return self._parse_book_snapshot(record, tick_size_pct=tick_size_pct, depth=depth)

        if data_type == "trade":
            record["side"] = 1 if record["side"] == "buy" else -1

        column_names = self._get_column_names(data_type)
        values = [record.get(col) for col in column_names]
        return values

    def _parse_book_snapshot(self, record: dict, tick_size_pct: float = 0.01, depth: int = 100) -> list | None:
        """
        Parse a book snapshot record from the normalized API response into a list of values
        """
        if tick_size_pct == 0:
            raise ValueError("tick_size_pct=0 is not supported yet")

        bids = record.get("bids", [])
        asks = record.get("asks", [])
        if not bids or not asks:
            return None

        # bids are of type [{"price": float, "size": float}, ...]
        # turn them to bids array of shape (depth, 2)
        bids = np.array([[bid["price"], bid["amount"]] for bid in bids if "price" in bid and "amount" in bid])
        asks = np.array([[ask["price"], ask["amount"]] for ask in asks if "price" in ask and "amount" in ask])

        top_bid, top_ask = bids[0][0], asks[0][0]
        mid_price = (top_bid + top_ask) / 2
        tick_size = mid_price * tick_size_pct / 100

        raw_bids = np.zeros(depth, dtype=np.float64)
        raw_asks = np.zeros(depth, dtype=np.float64)

        top_bid, _bids = accumulate_orderbook_levels(bids, raw_bids, tick_size, True, depth, False)
        top_ask, _asks = accumulate_orderbook_levels(asks, raw_asks, tick_size, False, depth, False)

        return [record["localTimestamp"], top_bid, top_ask, tick_size, _bids, _asks]

    def get_names(self, exchange: str | None = None, data_type: str | None = None) -> list[str]:
        """
        Get available data names (symbols) for the specified exchange and data type

        Parameters:
        -----------
        exchange : str | None
            Optional exchange name to filter symbols
        data_type : str | None
            Optional data type to filter symbols (e.g., 'trades', 'book')

        Returns:
        --------
        list[str]
            List of available data names in the format 'exchange:symbol'
        """
        if not exchange:
            exchanges = self.get_exchanges()
            symbols = []
            for exch in exchanges:
                symbols.extend(self.get_names(exch, data_type))
            return symbols

        # Get exchange info to get available symbols
        exchange_info = self.get_exchange_info(exchange)
        if not exchange_info or "availableSymbols" not in exchange_info:
            return []

        symbols = []
        for symbol_info in exchange_info["availableSymbols"]:
            symbol_id = symbol_info["id"]

            # Filter by data type if specified
            if data_type and "availableChannels" in exchange_info:
                if data_type not in exchange_info["availableChannels"]:
                    continue

            symbols.append(f"{exchange}:{symbol_id}")

        return symbols

    def _build_chunks(
        self,
        line_queue: Queue,
        chunk_queue: Queue,
        data_type: str,
        tick_size_pct: float,
        depth: int,
        start: pd.Timestamp,
        stop_event: threading.Event,
        chunksize: int,
        data_id: str,
    ):
        """Build chunks from the raw data stream asynchronously"""
        current_chunk = []
        current_type = None
        prev_record_time = None

        try:
            while not stop_event.is_set():
                try:
                    line = line_queue.get(timeout=30.0)  # Increase timeout for better reliability
                    if line is None:  # End of stream
                        break
                except Exception as e:  # Queue.Empty or other errors
                    logger.warning(f"Timeout or error getting line in _build_chunks for {data_id}: {e}")
                    continue

                # Parse the JSON object
                try:
                    record = self._parse_line(line)
                    if not record:
                        continue
                    # Skip disconnect messages unless specifically requested
                    if record.get("type") == "disconnect" and "disconnect" not in data_type:
                        continue
                    if "localTimestamp" not in record:
                        continue

                    # Get the record type
                    record_type = record.get("type", "")

                    # For the first valid record, set the record type
                    if record_type and not current_type:
                        current_type = record_type

                    # Only process records of the same type
                    if record_type == current_type and current_type:
                        # Skip records before actual start time for book snapshot data
                        if current_type == "book_snapshot":
                            record_time = pd.Timestamp(record["localTimestamp"])
                            if prev_record_time is None or record_time.floor("h") != prev_record_time.floor("h"):
                                prev_record_time = record_time
                                logger.info(f"{data_id} New hour: {record_time}")
                            if record_time < start:
                                continue

                        # Parse the record into a list of values
                        values = self._parse_record(record, current_type, tick_size_pct=tick_size_pct, depth=depth)
                        if values:
                            current_chunk.append(values)

                            # If we have enough records for a chunk, put it in the queue
                            if len(current_chunk) >= chunksize:
                                try:
                                    # Use put with timeout to avoid blocking indefinitely
                                    chunk_queue.put(current_chunk)
                                    current_chunk = []
                                except Exception as e:
                                    logger.error(f"Error putting chunk on queue for {data_id}: {e}")
                                    # If we can't put the chunk, we should stop
                                    stop_event.set()
                                    break
                except Exception as e:
                    logger.error(f"Error processing record in _build_chunks for {data_id}: {e}")
                    continue

            # Put any remaining records
            if current_chunk and not stop_event.is_set():
                try:
                    chunk_queue.put(current_chunk)
                except Exception as e:
                    logger.error(f"Error putting final chunk on queue for {data_id}: {e}")
        except Exception as e:
            logger.error(f"Unexpected error in _build_chunks for {data_id}: {e}")
        finally:
            # Always signal end of chunks, even in error case
            try:
                chunk_queue.put(None)
            except:
                pass  # If this fails, we can't do much about it

    def read(
        self,
        data_id: str,
        start: str | None = None,
        stop: str | None = None,
        transform: DataTransformer = DataTransformer(),
        chunksize=0,
        timeframe=None,
        data_type="trade",
        tick_size_pct: float = 0.01,
        depth: int = 100,
        **kwargs,
    ) -> Iterable | Any:
        try:
            exchange, symbol = data_id.split(":", 1)
        except ValueError:
            raise ValueError(f"Invalid data_id format: {data_id}. Expected format: 'exchange:symbol'")

        exchange = TARDIS_EXCHANGE_MAPPERS.get(exchange.lower(), exchange)
        symbol = self._map_symbol(symbol, exchange)

        start_date, end_date = handle_start_stop(start, stop)
        if not start_date:
            raise ValueError("Start date must be provided for TardisMachineReader")

        # Handle start and stop dates
        actual_start_date = pd.Timestamp(start_date)
        start_date = (
            actual_start_date.floor("d") if data_type.startswith(("book_snapshot", "orderbook")) else actual_start_date
        )
        end_date = pd.Timestamp(stop) if stop else (actual_start_date + pd.Timedelta(days=1))

        # latest 15min of data is not available
        _now = pd.Timestamp(time_now())
        _now_offset = pd.Timedelta("15min")
        if _now - pd.Timestamp(end_date) < _now_offset:
            actual_start_date -= _now_offset
            end_date = _now - _now_offset

        assert isinstance(actual_start_date, pd.Timestamp)
        assert isinstance(start_date, pd.Timestamp)
        assert isinstance(end_date, pd.Timestamp)

        # Determine the data types to request
        data_types = self._get_normalized_data_type(data_type, timeframe)

        # Format the request for the normalized API
        options = {
            "exchange": exchange,
            "symbols": [symbol],
            "from": start_date.isoformat(),
            "to": end_date.isoformat(),
            "dataTypes": data_types,
            "withDisconnectMessages": False,
        }

        # Construct the URL with options as a query parameter
        url = f"{self.machine_url}/replay-normalized?options={json.dumps(options)}"

        logger.debug(f"Requesting data from Tardis Machine normalized API: {url}")

        # Setup streaming
        line_queue = Queue()
        stop_event = threading.Event()

        try:
            stream_future = self._stream_data(url, line_queue, stop_event)

            if chunksize > 0:
                # Setup chunk processing
                chunk_queue = Queue(maxsize=10)

                # Create a daemon thread to avoid hanging on program exit
                chunk_builder = threading.Thread(
                    target=self._build_chunks,
                    args=(
                        line_queue,
                        chunk_queue,
                        data_type,
                        tick_size_pct,
                        depth,
                        actual_start_date,
                        stop_event,
                        chunksize,
                        data_id,
                    ),
                    daemon=True,  # Mark as daemon thread
                )
                chunk_builder.start()

                # For chunked processing, wrap the iterator in a cleanup generator
                def chunked_stream():
                    error_encountered = False
                    try:
                        while not stop_event.is_set():
                            try:
                                chunk = chunk_queue.get()
                                if chunk is None:  # End of chunks
                                    logger.info(f"End of chunks for {data_id}")
                                    break

                                # Process and yield the chunk
                                transform.start_transform(
                                    data_id,
                                    self._get_column_names(data_type),
                                    start=start_date.isoformat(),
                                    stop=end_date.isoformat(),
                                )
                                transform.process_data(chunk)
                                yield transform.collect()
                            except Exception as e:
                                logger.error(f"Error processing chunk for {data_id}: {e}")
                                error_encountered = True
                                break
                    finally:
                        # Always ensure cleanup
                        stop_event.set()

                        # Wait for streaming to complete, but with a timeout
                        try:
                            stream_future.result(timeout=10)
                        except Exception as e:
                            logger.warning(f"Error or timeout waiting for stream to complete: {e}")

                        # Only wait for the chunk builder if it's still alive and we didn't hit an error
                        if not error_encountered and chunk_builder.is_alive():
                            chunk_builder.join(timeout=10)

                return chunked_stream()
            else:
                # For non-chunked processing, we can use try/finally directly
                try:
                    return self._process_stream(
                        line_queue, data_id, transform, data_type, tick_size_pct, depth, actual_start_date, end_date
                    )
                finally:
                    stop_event.set()
                    try:
                        stream_future.result(timeout=10)  # Add timeout to prevent hanging
                    except Exception as e:
                        logger.warning(f"Error or timeout waiting for stream to complete: {e}")
        except Exception as e:
            logger.error(f"Error in read operation for {data_id}: {e}")
            stop_event.set()  # Ensure stop event is set in case of exceptions
            raise

    def _process_stream(
        self,
        line_queue: Queue,
        data_id: str,
        transform: DataTransformer,
        data_type: str,
        tick_size_pct: float,
        depth: int,
        start: pd.Timestamp,
        stop: pd.Timestamp,
    ):
        """Process the entire stream at once"""
        all_records = []
        current_type = None
        column_names = []

        while True:
            try:
                line = line_queue.get(timeout=60)  # Add timeout to prevent hanging
                if line is None:  # End of stream
                    break
            except:  # Queue.Empty or other errors
                logger.warning(f"Timeout waiting for data in _process_stream for {data_id}")
                break

            # Parse the JSON object
            record = self._parse_line(line)
            if not record:
                continue

            # Skip disconnect messages unless specifically requested
            if record.get("type") == "disconnect" and "disconnect" not in data_type:
                continue

            if "localTimestamp" not in record:
                continue

            record_time = pd.Timestamp(record["localTimestamp"])
            if record_time < start:
                continue

            # Get the record type
            current_type = record.get("type", "")

            # For the first valid record, set the record type and column names
            if current_type and not column_names:
                column_names = self._get_column_names(current_type)

            # Only process records of the same type
            if current_type:
                # Parse the record into a list of values
                values = self._parse_record(record, current_type, tick_size_pct=tick_size_pct, depth=depth)
                if values:
                    all_records.append(values)

        # Process all records at once
        if all_records:
            transform.start_transform(data_id, column_names, start=start.isoformat(), stop=stop.isoformat())
            transform.process_data(all_records)
            return transform.collect()
        else:
            logger.warning("No valid records found in Tardis Machine normalized response")
            return None

    def get_exchanges(self) -> list[str]:
        """
        Get list of available exchanges from Tardis.dev API

        Returns:
        --------
        list[str]
            List of exchange IDs
        """
        if self._exchanges_cache is not None:
            return self._exchanges_cache

        try:
            url = f"{self.api_url}/exchanges"
            status, data = self._async_loop.submit(self._http_get(url)).result()

            if status != 200:
                logger.warning(f"Failed to get exchanges: {data}")
                return []

            self._exchanges_cache = [exchange["id"] for exchange in data]
            return self._exchanges_cache

        except Exception as e:
            logger.error(f"Error getting exchanges: {str(e)}")
            return []

    def get_exchange_info(self, exchange: str) -> dict | None:
        """
        Get detailed information about an exchange from Tardis.dev API

        Parameters:
        -----------
        exchange : str
            Exchange ID

        Returns:
        --------
        dict | None
            Dictionary containing exchange information or None if not found
        """
        exchange = TARDIS_EXCHANGE_MAPPERS.get(exchange.lower(), exchange)
        if exchange in self._exchange_info_cache:
            return self._exchange_info_cache[exchange]

        try:
            url = f"{self.api_url}/exchanges/{exchange}"
            status, data = self._async_loop.submit(self._http_get(url)).result()

            if status != 200:
                logger.warning(f"Failed to get exchange info for {exchange}: {data}")
                return None

            self._exchange_info_cache[exchange] = data
            return data

        except Exception as e:
            logger.error(f"Error getting exchange info for {exchange}: {str(e)}")
            return None

    def get_data_types(self, exchange: str) -> list[str]:
        """
        Get available data types (channels) for an exchange

        Parameters:
        -----------
        exchange : str
            Exchange ID

        Returns:
        --------
        list[str]
            List of available data types
        """
        exchange_info = self.get_exchange_info(exchange)
        if not exchange_info or "availableChannels" not in exchange_info:
            return []

        return exchange_info["availableChannels"]

    def get_symbols(self, exchange: str, dtype: str) -> list[str]:
        """
        Get symbols for a specific exchange and data type

        Parameters:
        -----------
        exchange : str
            Exchange ID
        dtype : str
            Data type

        Returns:
        --------
        list[str]
            List of symbols in the format 'exchange:symbol'
        """
        exchange_info = self.get_exchange_info(exchange)
        if not exchange_info or "availableSymbols" not in exchange_info:
            return []

        # Check if the data type is available for this exchange
        if dtype and "availableChannels" in exchange_info:
            if dtype not in exchange_info["availableChannels"]:
                return []

        return [f"{exchange}:{symbol['id']}" for symbol in exchange_info["availableSymbols"]]

    def get_time_ranges(self, symbol: str, dtype: str) -> tuple[np.datetime64, np.datetime64]:
        """
        Get available time range for a symbol and data type

        Parameters:
        -----------
        symbol : str
            Symbol in the format 'exchange:symbol'
        dtype : str
            Data type

        Returns:
        --------
        tuple[np.datetime64, np.datetime64]
            Tuple of (start_time, end_time)
        """
        try:
            exchange, symbol_id = symbol.split(":")
            exchange_info = self.get_exchange_info(exchange)
            if not exchange_info or "availableSymbols" not in exchange_info:
                return np.datetime64("NaT"), np.datetime64("NaT")

            # Find the symbol in the available symbols
            for symbol_info in exchange_info["availableSymbols"]:
                if symbol_info["id"].lower() == symbol_id.lower():
                    # Get the availability dates
                    available_since = symbol_info.get("availableSince")
                    if available_since:
                        start_time = np.datetime64(available_since)
                        # End time is typically the current date
                        end_time = np.datetime64(pd.Timestamp.now().date())
                        return start_time, end_time

            return np.datetime64("NaT"), np.datetime64("NaT")

        except Exception as e:
            logger.error(f"Error getting time ranges for {symbol}: {str(e)}")
            return np.datetime64("NaT"), np.datetime64("NaT")

    def _parse_line(self, line: str) -> dict | None:
        try:
            record = orjson.loads(line)
            for col in ["timestamp", "localTimestamp"]:
                record[col] = record[col].rstrip("Z")
            return record
        except Exception as _:
            logger.warning(f"Failed to parse line as JSON: {line[:100]}...")
            return None

    def _stream_data(self, url: str, line_queue: Queue, stop_event: threading.Event):
        """Submit the streaming coroutine to the asyncio loop"""
        return self._async_loop.submit(self._fetch_stream(url, line_queue, stop_event))

    def _map_symbol(self, symbol: str, exchange: str) -> str:
        """Map symbol to Tardis Machine API format"""
        if exchange.lower() == "bitfinex-derivatives":
            return f"{symbol[:3]}F0:USTF0"
        else:
            return symbol



from typing import Literal

import matplotlib.colors as mcolors
import matplotlib.pyplot as plt
import pandas as pd
import plotly.graph_objects as go
from plotly.graph_objs.graph_objs import FigureWidget
from plotly.subplots import make_subplots

from qubx.core.series import OHLCV, TimeSeries
from qubx.utils import Struct
from qubx.utils.charting.mpl_helpers import ohlc_plot, plot_trends, subplot


def install_plotly_helpers():
    try:

        def rline_(look, x, y, c="red", lw=1):
            """
            Ray line
            """
            return look.update_layout(
                shapes=(
                    dict(
                        type="line",
                        xref="x1",
                        yref="y1",
                        x0=pd.Timestamp(x),
                        y0=y,
                        x1=look.data[0]["x"][-1],
                        y1=y,
                        fillcolor=c,
                        opacity=1,
                        line=dict(color=c, width=lw),
                    )
                ),
                overwrite=False,
            )

        def rline(look, x, y, c="red", lw=1, ls=None):
            return look.add_shape(
                go.layout.Shape(
                    type="line",
                    x0=pd.Timestamp(x),
                    x1=look.data[0]["x"][-1],
                    y0=y,
                    y1=y,
                    xref="x1",
                    yref="y1",
                    line=dict(width=lw, color=c, dash=ls),
                )
            )

        def rlinex(look, x0, x1, y, c="red", lw=1, ls=None):
            return look.add_shape(
                go.layout.Shape(
                    type="line",
                    x0=pd.Timestamp(x0),
                    x1=pd.Timestamp(x1),
                    y0=y,
                    y1=y,
                    xref="x1",
                    yref="y1",
                    line=dict(width=lw, color=c, dash=ls),
                )
            )

        def dliney(look, x0, y0, y1, c="red", lw=1, ls=None):
            return look.add_shape(
                go.layout.Shape(
                    type="line",
                    x0=pd.Timestamp(x0),
                    x1=pd.Timestamp(x0),
                    y0=y0,
                    y1=y1,
                    xref="x1",
                    yref="y1",
                    line=dict(width=lw, color=c, dash=ls),
                )
            )

        def vline(look, x, c="yellow", lw=1, ls="dot"):
            return look.add_shape(
                go.layout.Shape(
                    type="line",
                    x0=pd.Timestamp(x),
                    x1=pd.Timestamp(x),
                    y0=0,
                    y1=1,
                    xref="x1",
                    yref="paper",
                    line=dict(width=lw, color=c, dash=ls),
                )
            )

        def hline(look, y, c="yellow", lw=1, ls="dot"):
            return look.add_shape(
                go.layout.Shape(
                    type="line",
                    x0=0,
                    x1=1,
                    y0=y,
                    y1=y,
                    xref="paper",
                    yref="y1",
                    line=dict(width=lw, color=c, dash=ls),
                )
            )

        def arrow(look, x2, y2, x1, y1, c="red", text="", lw=1, font=dict(size=8), head=1):
            return look.add_annotation(
                x=x1,
                y=y1,
                ax=x2,
                ay=y2,
                xref="x",
                yref="y",
                axref="x",
                ayref="y",
                text=text,
                font=font,
                showarrow=True,
                arrowhead=head,
                arrowsize=1,
                arrowwidth=lw,
                arrowcolor=c,
            )

        def hover(v, h=600, n=2, legend=False, show_info=True, w=None):
            return (
                v.update_traces(xaxis="x1")
                .update_layout(
                    height=h,
                    width=w,
                    hovermode="x unified",
                    showlegend=legend,
                    hoverdistance=1 if show_info else 0,
                    xaxis={"hoverformat": "%d-%b-%y %H:%M"},
                    yaxis={"hoverformat": f".{n}f"},
                    dragmode="zoom",
                    newshape=dict(line_color="yellow", line_width=1.0),
                    modebar_add=["drawline", "drawopenpath", "drawrect", "eraseshape"],
                    # hoversubplots="axis",
                    hoverlabel=dict(align="auto", bgcolor="rgba(10, 10, 10, 0.5)"),
                )
                .update_xaxes(
                    showspikes=True,
                    spikemode="across",
                    spikesnap="cursor",
                    spikecolor="#306020",
                    spikethickness=1,
                    spikedash="dot",
                )
                .update_yaxes(
                    spikesnap="cursor",
                    spikecolor="#306020",
                    tickformat=f".{n}f",
                    spikethickness=1,
                )
            )

        def to_obsidian_json(f: FigureWidget, file_nm: str | None) -> str | FigureWidget:
            """
            Convert figure to be pasted into Obsidian md file
            """
            import datetime
            import json

            import numpy as np

            ss1 = f.to_plotly_json()
            r = {}
            for k, v in ss1.items():
                if k == "data":
                    rl = []
                    for ds in v:
                        kk = {}
                        for dk, dv in ds.items():
                            if isinstance(dv, np.ndarray):
                                if isinstance(dv[0], datetime.date):
                                    kk[dk] = [x.isoformat() for x in dv]
                                else:
                                    kk[dk] = list(dv)
                            else:
                                kk[dk] = dv
                        rl.append(kk)
                    r[k] = rl
                else:
                    r[k] = v
            s = f"```plotly fold\n\n\n\n\n{json.dumps(r, indent=1)}\n```"

            if file_nm:
                from pathlib import Path

                with open(Path(file_nm).expanduser().with_suffix(".md"), "w") as file:
                    file.write("---\nsticker: lucide//line-chart\n---\n\n" + s)

                return f

            return s

        FigureWidget.hover = hover  # type: ignore
        FigureWidget.rline = rline  # type: ignore
        FigureWidget.rlinex = rlinex  # type: ignore
        FigureWidget.rline_ = rline_  # type: ignore
        FigureWidget.vline = vline  # type: ignore
        FigureWidget.hline = hline  # type: ignore
        FigureWidget.dliney = dliney  # type: ignore
        FigureWidget.arrow = arrow  # type: ignore
        FigureWidget.to_obsidian_json = to_obsidian_json  # type: ignore

    except:  # noqa: E722
        print(" >>> Cant attach helpers to plotly::FigureWidget - probably it isn't installed !")


# - install plotly helpers
install_plotly_helpers()


class AbstractLookingGlass:
    """
    Handy utility for plotting data
    """

    def __init__(self, master, studies: dict | None, title=""):
        self.m = master
        self.s = {} if studies is None else studies
        self._title = title

    def look(self, *args, title=None, **kwargs):
        _vert_bar = None
        zoom = None

        if len(args) == 1:
            zoom = args[0]
        elif len(args) > 1:
            zoom = args

        if zoom and not isinstance(zoom, slice):
            zoom = zoom if isinstance(zoom, (list, tuple)) else [zoom]

            if len(zoom) == 2:
                z0, is_d_0 = self.__as_time(zoom[0])
                z1, is_d_1 = self.__as_time(zoom[1])

                if is_d_0:
                    if is_d_1:
                        raise ValueError("At least one of zoom values must be timestamp !")
                    zoom = slice(z1 - z0, z1)
                else:
                    zoom = slice(z0, (z0 + z1) if is_d_1 else z1)
            elif len(zoom) > 2:
                z0, is_d_0 = self.__as_time(zoom[0])
                z1, is_d_1 = self.__as_time(zoom[1])
                z2, is_d_2 = self.__as_time(zoom[2])

                if is_d_1:
                    raise ValueError("Second argument must be timestamp !")
                if not is_d_0:
                    raise ValueError("First argument must be timedelta !")
                if not is_d_2:
                    raise ValueError("Third argument must be timedelta !")

                zoom = slice(z1 - z0, z1 + z2)
                _vert_bar = z1
            elif len(zoom) == 1:
                import datetime

                z1, is_d = self.__as_time(zoom[0])
                if is_d:
                    raise ValueError("Argument must be date time not timedelta !")

                if z1.time() == datetime.time(0, 0):
                    shift = pd.Timedelta(kwargs.get("shift", "24h"))
                    zoom = slice(z1, z1 + shift)
                else:
                    shift = pd.Timedelta(kwargs.get("shift", "12h"))
                    zoom = slice(z1 - shift, z1 + shift)
                    _vert_bar = z1
            else:
                raise ValueError("Don't know how to interpret '%s'" % str(zoom))

        return self._show_plot(_vert_bar, title, zoom)

    def _frame_has_cols(self, df, cols):
        return isinstance(df, pd.DataFrame) and all(x in df.columns for x in cols)

    def __as_time(self, z):
        _is_delta = False
        if isinstance(z, str):
            try:
                z = pd.Timedelta(z)
                _is_delta = True
            except:
                try:
                    z = pd.Timestamp(z)
                except:
                    raise ValueError("Value '%s' can't be recognized" % z)
        else:
            _is_delta = isinstance(z, pd.Timedelta)
        return z, _is_delta

    def _show_plot(self, _vert_bar, title, zoom):
        raise NotImplementedError("Must be implemented in child class %s", self.__class__.__name__)


class LookingGlass:
    def __init__(
        self,
        master,
        studies: dict | None = None,
        title="",
        backend: Literal["matplotlib", "mpl", "plotly", "ply", "plt"] = "plotly",
        **kwargs,
    ):
        if backend in ["matplotlib", "mpl"]:
            self.__instance = LookingGlassMatplotLib(master=master, studies=studies, title=title, **kwargs)
        elif backend in ["plotly", "ply", "plt"]:
            self.__instance = LookingGlassPlotly(master=master, studies=studies, title=title, **kwargs)
        else:
            raise ValueError("Backend %s is not recognized" % backend)

    def look(self, *args, **kwargs):
        return self.__instance.look(*args, **kwargs)


class LookingGlassMatplotLib(AbstractLookingGlass):
    def __init__(
        self,
        master,
        studies: dict | None = None,
        master_size=3,
        study_size=1,
        title="",
        legend_loc="upper left",
        fmt="%H:%M",
        ohlc_width=0,
    ):
        super().__init__(master, studies, title)
        self.s_size = study_size
        self.m_size = master_size
        self.legend_loc = legend_loc
        self._fmt = fmt
        self._ohlc_width = ohlc_width
        self._n_style = "-"
        self._last_color = None
        self._last_mark = None

    def _plt_series(self, y, zoom, study_name, k, plot_style="line"):
        _forced_limits = []
        if isinstance(y, (int, float)):
            try:
                plt.axhline(y, lw=0.5, ls=self._n_style)
            except:
                plt.axhline(y, lw=0.5, ls="--")
        else:
            _lbl = y.name if hasattr(y, "name") and y.name else ("%s_%d" % (study_name, k))

            if isinstance(y, (pd.DataFrame, OHLCV)):
                y = y.pd() if isinstance(y, OHLCV) else y

                yy = y[zoom] if zoom else y

                # reversal points
                if self._frame_has_cols(y, ["start_price", "delta"]):
                    _lo_pts = yy[yy.delta > 0].rename(columns={"start_price": "Bottom"})
                    _hi_pts = yy[yy.delta < 0].rename(columns={"start_price": "Top"})
                    plt.plot(
                        _lo_pts.index,
                        _lo_pts.Bottom,
                        marker=6,
                        markersize=5,
                        ls="",
                        c="#10aa10",
                    )
                    plt.plot(_hi_pts.index, _hi_pts.Top, marker=7, markersize=5, ls="", c="r")

                # trends
                elif self._frame_has_cols(y, ["UpTrends", "DownTrends"]):
                    plot_trends(yy, fmt=self._fmt)

                # tracks
                elif self._frame_has_cols(y, ["Type", "Time", "Price", "PriceOccured"]):
                    _bot = yy[yy.Type == "-"]
                    _top = yy[yy.Type == "+"]
                    plt.plot(
                        _bot.index,
                        _bot.PriceOccured.rename("B"),
                        marker=5,
                        markersize=8,
                        ls="",
                        c="#1080ff",
                    )
                    plt.plot(
                        _top.index,
                        _top.PriceOccured.rename("T"),
                        marker=5,
                        markersize=8,
                        ls="",
                        c="#909010",
                    )

                # executed signals from signal tester
                elif self._frame_has_cols(y, ["exec_price", "quantity"]):
                    _b_ords = yy[yy.quantity > 0]
                    _s_ords = yy[yy.quantity < 0]
                    plt.plot(
                        _b_ords.index,
                        _b_ords.exec_price.rename("BOT"),
                        marker="2",
                        markersize=10,
                        ls="",
                        c="#1080ff",
                    )
                    plt.plot(
                        _s_ords.index,
                        _s_ords.exec_price.rename("SLD"),
                        marker="1",
                        markersize=10,
                        ls="",
                        c="#909010",
                    )

                # order executions from simulator
                elif self._frame_has_cols(y, ["side", "fill_avg_price", "quantity", "status"]):
                    _b_ords = yy[yy.side == "BUY"]
                    _s_ords = yy[yy.side == "SELL"]
                    plt.plot(
                        _b_ords.index,
                        _b_ords.fill_avg_price.rename("BOT"),
                        marker="2",
                        markersize=10,
                        ls="",
                        c="#1080ff",
                    )
                    plt.plot(
                        _s_ords.index,
                        _s_ords.fill_avg_price.rename("SLD"),
                        marker="1",
                        markersize=10,
                        ls="",
                        c="#909010",
                    )

                # experimental tester view of trading log
                elif self._frame_has_cols(y, ["Action", "Price", "Info", "PnL_ticks"]):
                    _b_ords = yy[yy.Action == "Long"]
                    _s_ords = yy[yy.Action == "Short"]
                    _t_ords = yy[yy.Action == "Take"]
                    _l_ords = yy[yy.Action == "Stop"]
                    _e_ords = yy[(yy.Action == "Expired") | (yy.Action == "Flat")]
                    plt.plot(
                        _b_ords.index,
                        _b_ords.Price.rename("BOT"),
                        marker=6,
                        markersize=10,
                        ls="",
                        c="#3cfa00",
                    )
                    plt.plot(
                        _s_ords.index,
                        _s_ords.Price.rename("SLD"),
                        marker=7,
                        markersize=10,
                        ls="",
                        c="#20ffff",
                    )
                    plt.plot(
                        _t_ords.index,
                        _t_ords.CurrentPrice.rename("Take"),
                        marker="P",
                        markersize=10,
                        ls="",
                        c="#fffb00",
                    )
                    plt.plot(
                        _l_ords.index,
                        _l_ords.CurrentPrice.rename("Stop"),
                        marker="8",
                        markersize=10,
                        ls="",
                        c="#fffb00",
                    )
                    plt.plot(
                        _e_ords.index,
                        _e_ords.CurrentPrice.rename("Exp"),
                        marker="X",
                        markersize=10,
                        ls="",
                        c="#b0b0b0",
                    )

                elif self._frame_has_cols(y, ["open", "high", "low", "close"]):
                    ohlc_plot(yy, width=self._ohlc_width, fmt=self._fmt)
                    _forced_limits = 0.999 * min(yy["low"]), 1.001 * max(yy["high"])
                    # temp hack to aling scales
                    plt.plot(yy["close"], lw=0, label=_lbl)
                else:
                    for _col in yy.columns:
                        self._plot_as_type(yy[_col], plot_style, self._n_style, _col)
            else:
                y = y.pd() if isinstance(y, TimeSeries) else y
                yy = y[zoom] if zoom else y
                self._plot_as_type(yy, plot_style, self._n_style, _lbl)

            # we want to see OHLC at maximal scale
            return _forced_limits

    def _extract_colors_marks(self, line: str):
        from matplotlib.markers import MarkerStyle

        mplc = mcolors.get_named_colors_mapping()

        colors, marks = [], []
        r_l = ""
        for c in (line if line else "").split(" "):
            if c.startswith("#") or c.lower() in mplc:
                colors.append(c)
                continue
            if c.startswith("arrow-") or c in MarkerStyle.markers:
                marks.append("^" if c.endswith("arrow-up") else "v" if c.endswith("arrow-down") else c)
                continue
            if c.startswith("dots") or c in MarkerStyle.markers:
                marks.append(".")
                continue
            r_l += c
        return r_l if r_l else None, colors, marks

    def _plot_as_type(self, y, plot_style, line_style: str, label):
        _l_style, clrs, mrks = self._extract_colors_marks(line_style)
        _clr = clrs[0] if clrs else None
        if _clr:
            self._last_color = _clr

        if plot_style == "line":
            _l_style = "" if self._last_mark else _l_style
            plt.plot(
                y,
                ls=_l_style,
                label=label,
                color=self._last_color,
                marker=self._last_mark,
                ms=1 if self._last_mark == "." else 6,
            )

        elif plot_style == "area":
            plt.fill_between(y.index, y, color=_clr, label=label)
        elif plot_style.startswith("step"):
            _where = "post" if "post" in plot_style else "pre"
            plt.step(y.index, y, color=_clr, where=_where, label=label)
        elif plot_style.startswith("bar"):
            _bw = pd.Series(y.index).diff().mean().total_seconds() / 24 / 60 / 60  # type: ignore
            plt.bar(y.index, y, lw=0.4, width=_bw, edgecolor=_clr, color=_clr, label=label)

        # - reset last mark
        self._last_mark = None

    def _show_plot(self, vert_bar, title, zoom):
        # plot all master series
        shape = (self.s_size * len(self.s) + self.m_size, 1)
        subplot(shape, 1, rowspan=self.m_size)
        ms = self.m if isinstance(self.m, (tuple, list)) else [self.m]
        _limits_to_set = None

        for j, m in enumerate(ms):
            # if style description
            if isinstance(m, str):
                _l_style, clrs, mrks = self._extract_colors_marks(m)
                _clr = clrs[0] if clrs else None
                _marker = mrks[0] if mrks else None
                self._n_style = _l_style
                self._last_color = _clr if _clr else self._last_color
                self._last_mark = _marker if _marker else self._last_mark

            else:
                _lims = self._plt_series(m, zoom, "Master", j)
                self._n_style = "-"
                if _limits_to_set is None and _lims:
                    _limits_to_set = _lims

        # special case
        if _limits_to_set:
            plt.ylim(*_limits_to_set)

        if self.legend_loc:
            plt.legend(loc=self.legend_loc)

        if vert_bar:
            plt.axvline(vert_bar, ls="-.", lw=0.5)
            if title is None:
                plt.title("%s %s" % (self._title, str(vert_bar)))

        if title is not None:
            plt.title("%s %s" % (self._title, str(title)))

        # plot studies
        i = 1 + self.m_size
        for k, vs in self.s.items():
            subplot(shape, i, rowspan=self.s_size)
            vs = vs if isinstance(vs, (tuple, list)) else [vs]
            self._n_style = "-"
            plot_style = "line"

            wait_for_limits = False
            for j, v in enumerate(vs):
                # if we need to read limits
                if wait_for_limits:
                    if isinstance(v, (list, tuple)) and len(v) > 1:
                        plt.ylim(*v)
                    wait_for_limits = False
                    continue

                # if style description
                if isinstance(v, str):
                    vl = v.lower()
                    if vl.startswith("lim"):
                        wait_for_limits = True
                    elif any([vl.startswith(x) for x in ["line", "bar", "step", "stem", "area"]]):
                        plot_style = vl
                    else:
                        self._n_style = v
                else:
                    self._plt_series(v, zoom, k, j, plot_style=plot_style)

            if vert_bar:
                plt.axvline(vert_bar, ls="-.", lw=0.5)

            i += self.s_size
            plt.legend(loc=self.legend_loc)

        self._n_style = "-"

        class EmptyPlot:
            def hover(self, *args, **kwargs):
                dpi = plt.rcParams["figure.dpi"]

                h = kwargs.get("h") or kwargs.get("height")
                w = kwargs.get("w") or kwargs.get("width")

                if h:
                    h = max(1, h / dpi)
                if w:
                    w = max(1, w / dpi)

                if h and not w:
                    w = 1.618 * h
                elif w and not h:
                    h = 0.618 * w

                if h and w:
                    plt.gcf().set_size_inches(w, h)

                return self

            def show(self, *args, **kwargs):
                pass

            def __str__(self):
                return title if title else ""

            def __repr__(self):
                return title if title else ""

        return EmptyPlot()


class LookingGlassPlotly(AbstractLookingGlass):
    TREND_COLORS = Struct(
        uline="#ffffff",
        dline="#ffffff",
        udot="rgb(10,168,10)",
        ddot="rgb(168,10,10)",
    )

    def __init__(
        self,
        master,
        studies: dict | None = None,
        master_plot_height=400,
        study_plot_height=100,
        title="",
    ):
        super().__init__(master, studies, title)
        self.mph = master_plot_height
        self.sph = study_plot_height
        self._n_style = "-"

    def __plt_series(self, y, zoom, study_name, k, row, col, plot_style="line"):
        _lbl = y.name if hasattr(y, "name") and y.name else ("%s_%d" % (study_name, k))

        def _scatter(
            xs: pd.Series, comments: pd.Series | None, name: str, marker: str, color: str, size: int = 11
        ) -> None:
            _args = dict(
                x=xs.index,
                y=xs,
                mode="markers",
                name=name,
                text=comments,
                marker={
                    "symbol": marker,
                    "size": size,
                    "color": color,
                },
            )
            self.fig.add_trace(go.Scatter(**_args), row=row, col=col)

        if isinstance(y, (pd.DataFrame, OHLCV)):
            y = y.pd() if isinstance(y, OHLCV) else y
            yy = y[zoom] if zoom else y

            # candlesticks
            if self._frame_has_cols(y, ["open", "high", "low", "close"]):
                self.fig.add_trace(
                    go.Candlestick(
                        x=yy.index,
                        open=yy["open"],
                        high=yy["high"],
                        low=yy["low"],
                        close=yy["close"],
                        name=_lbl,
                        line={"width": 1},
                    ),
                    row=row,
                    col=col,
                )

            # trends
            elif self._frame_has_cols(y, ["UpTrends", "DownTrends"]):
                u, d = yy.UpTrends.dropna(), yy.DownTrends.dropna()
                for i, r in enumerate(u.iterrows()):
                    self.fig.add_trace(
                        go.Scatter(
                            x=[r[0], r[1].end],
                            y=[r[1].start_price, r[1].end_price],
                            mode="lines+markers",
                            name="UpTrends",
                            line={
                                "color": LookingGlassPlotly.TREND_COLORS.uline,
                                "width": 1,
                                "dash": "dot",
                            },
                            marker={"color": LookingGlassPlotly.TREND_COLORS.udot},
                            showlegend=i == 0,
                            legendgroup="trends_UpTrends",
                        ),
                        row=row,
                        col=col,
                    )

                for i, r in enumerate(d.iterrows()):
                    self.fig.add_trace(
                        go.Scatter(
                            x=[r[0], r[1].end],
                            y=[r[1].start_price, r[1].end_price],
                            mode="lines+markers",
                            name="DownTrends",
                            line={
                                "color": LookingGlassPlotly.TREND_COLORS.dline,
                                "width": 1,
                                "dash": "dot",
                            },
                            marker={"color": LookingGlassPlotly.TREND_COLORS.ddot},
                            showlegend=i == 0,
                            legendgroup="trends_DownTrends",
                        ),
                        row=row,
                        col=col,
                    )

            # order executions from simulator
            elif self._frame_has_cols(y, ["side", "fill_avg_price", "quantity", "status"]):
                yy = yy[yy.status == "FILLED"]
                _b_ords = yy[yy.side == "BUY"]
                _s_ords = yy[yy.side == "SELL"]
                _info_b = "INFO : " + _b_ords["user_description"]
                _info_s = "INFO : " + _s_ords["user_description"]

                self.fig.add_trace(
                    go.Scatter(
                        x=_b_ords.index,
                        y=_b_ords.fill_avg_price,
                        mode="markers",
                        name="BOT",
                        text=_info_b,
                        marker={
                            "symbol": "triangle-up",
                            "size": 13,
                            "color": "#3cfa00",
                        },
                    ),
                    row=row,
                    col=col,
                )
                self.fig.add_trace(
                    go.Scatter(
                        x=_s_ords.index,
                        y=_s_ords.fill_avg_price,
                        mode="markers",
                        name="SLD",
                        text=_info_s,
                        marker={
                            "symbol": "triangle-down",
                            "size": 13,
                            "color": "#20ffff",
                        },
                    ),
                    row=row,
                    col=col,
                )

            # experimental tester view of trading log
            elif self._frame_has_cols(y, ["Action", "Price", "Info", "PnL_ticks"]):
                _b_ords = yy[yy.Action == "Long"]
                _s_ords = yy[yy.Action == "Short"]
                _t_ords = yy[yy.Action == "Take"]
                _l_ords = yy[yy.Action == "Stop"]
                _e_ords = yy[(yy.Action == "Expired") | (yy.Action == "Flat")]
                _info = "INFO : " + yy["Info"] + "<br>PnL Ticks: " + yy["PnL_ticks"].astype(str)

                self.fig.add_trace(
                    go.Scatter(
                        x=_b_ords.index,
                        y=_b_ords.Price,
                        mode="markers",
                        name="BOT",
                        text=_info,
                        marker={
                            "symbol": "triangle-up",
                            "size": 13,
                            "color": "#3cfa00",
                        },
                    ),
                    row=row,
                    col=col,
                )

                self.fig.add_trace(
                    go.Scatter(
                        x=_s_ords.index,
                        y=_s_ords.Price,
                        mode="markers",
                        name="SLD",
                        text=_info,
                        marker={
                            "symbol": "triangle-down",
                            "size": 13,
                            "color": "#20ffff",
                        },
                    ),
                    row=row,
                    col=col,
                )

                self.fig.add_trace(
                    go.Scatter(
                        x=_t_ords.index,
                        y=_t_ords.CurrentPrice,
                        mode="markers",
                        name="Take",
                        text=_info,
                        marker={"symbol": "cross", "size": 13},
                    ),
                    row=row,
                    col=col,
                )

                self.fig.add_trace(
                    go.Scatter(
                        x=_l_ords.index,
                        y=_l_ords.CurrentPrice,
                        mode="markers",
                        name="Stop",
                        text=_info,
                        marker={"symbol": "circle", "size": 13},
                    ),
                    row=row,
                    col=col,
                )

                self.fig.add_trace(
                    go.Scatter(
                        x=_e_ords.index,
                        y=_e_ords.CurrentPrice,
                        mode="markers",
                        name="Exp",
                        text=_info,
                        marker={"symbol": "x", "size": 13},
                    ),
                    row=row,
                    col=col,
                )

            # reversal points
            elif self._frame_has_cols(y, ["start_price", "delta"]):
                _lo_pts = yy[yy.delta > 0]
                _hi_pts = yy[yy.delta < 0]
                self.fig.add_trace(
                    go.Scatter(
                        x=_lo_pts.index,
                        y=_lo_pts.start_price,
                        mode="markers",
                        name="Bottom",
                        marker={"symbol": "triangle-up"},
                    ),
                    row=row,
                    col=col,
                )
                self.fig.add_trace(
                    go.Scatter(
                        x=_hi_pts.index,
                        y=_hi_pts.start_price,
                        mode="markers",
                        name="Top",
                        marker={"symbol": "triangle-down"},
                    ),
                    row=row,
                    col=col,
                )

            # tracks
            elif self._frame_has_cols(y, ["Type", "Time", "Price", "PriceOccured"]):
                _bot = yy[yy.Type == "-"]
                _top = yy[yy.Type == "+"]

                _scatter(_bot.PriceOccured, None, "B", "triangle-right", "#3cfa00", 12)
                _scatter(_top.PriceOccured, None, "T", "triangle-right", "#3cfa00", 12)

            # executions from executor's log
            elif self._frame_has_cols(y, ["exec_price", "quantity"]):
                _b_ords = yy[yy.quantity > 0]
                _s_ords = yy[yy.quantity < 0]

                # - how much was traded
                _b_info = "<i>bought</i> " + _b_ords.quantity.astype(str)
                _s_info = "<i>sold</i> " + _s_ords.quantity.astype(str)

                _scatter(_b_ords.exec_price, _b_info, "BOT", "triangle-up", "#25ff00", 12)
                _scatter(_s_ords.exec_price, _s_info, "SLD", "triangle-down", "#f0d", 12)
                # _scatter(_s_ords.exec_price, _s_info, "SLD", "triangle-down", "#20ffff", 12)

            # 27-aug-2024: show generated signals from Qubx backtester
            elif self._frame_has_cols(y, ["signal", "reference_price", "price", "take", "stop", "group", "comment"]):
                _sel_price = lambda x: x["reference_price"].where(x["price"].isna(), x["price"])  # noqa: E731
                _b_sigs = yy[yy.signal > 0]
                _z_sigs = yy[yy.signal == 0]
                _s_sigs = yy[yy.signal < 0]
                # - Longs
                _scatter(_sel_price(_b_sigs), _b_sigs["comment"], "LONG", "triangle-up-open-dot", "#3cfa00")
                _scatter(_b_sigs[~_b_sigs["take"].isna()]["take"], None, "Take", "line-ew-open", "#3cfa00")
                _scatter(_b_sigs[~_b_sigs["stop"].isna()]["stop"], None, "Stop", "line-ew-open", "#fa3c00")

                # - Shorts
                _scatter(_sel_price(_s_sigs), _s_sigs["comment"], "SHORTS", "triangle-down-open-dot", "#20ffff")
                _scatter(_s_sigs[~_s_sigs["take"].isna()]["take"], None, "Take", "line-ew-open", "#3cfa00")
                _scatter(_s_sigs[~_s_sigs["stop"].isna()]["stop"], None, "Stop", "line-ew-open", "#fa3c00")

                # - Exits
                _scatter(_sel_price(_z_sigs), _z_sigs["comment"], "EXITS", "circle-open-dot", "#ffffff")

            else:
                for _col in yy.columns:
                    self.__plot_as_type(yy[_col], row, col, plot_style, _col)
        else:
            y = y.pd() if isinstance(y, TimeSeries) else y
            yy = y[zoom] if zoom else y
            self.__plot_as_type(yy, row, col, plot_style, _lbl)

    def __plot_as_type(self, y, row, col, plot_style, label):
        style, color = self.__line_style_to_color_dash(self._n_style)
        if plot_style == "line":
            self.fig.add_trace(
                go.Scatter(
                    x=y.index, y=y, mode="lines", line={"width": 0.5, "dash": style, "color": color}, name=label
                ),
                row=row,
                col=col,
            )
        elif plot_style == "area":
            self.fig.add_trace(
                go.Scatter(x=y.index, y=y, mode="lines", line={"width": 1, "color": color}, fill="tozeroy", name=label),
                row=row,
                col=col,
            )
        elif plot_style.startswith("step"):
            self.fig.add_trace(
                go.Scatter(x=y.index, y=y, mode="lines", line={"shape": "hv", "color": color}, name=label),
                row=row,
                col=col,
            )
        elif plot_style.startswith("bar"):
            self.fig.add_trace(go.Bar(x=y.index, y=y, name=label, marker_color=color), row=row, col=col)
        elif plot_style.startswith("dots") or plot_style.startswith("point"):
            self.fig.add_trace(
                go.Scatter(
                    x=y.index,
                    y=y,
                    mode="markers",
                    name=label,
                    marker_color=color,
                    marker={"symbol": "circle", "size": 4},
                ),
                row=row,
                col=col,
            )
        elif plot_style.startswith("arrow"):
            self.fig.add_trace(
                go.Scatter(
                    x=y.index,
                    y=y,
                    mode="markers",
                    name=label,
                    marker_color=color,
                    marker={"symbol": plot_style, "size": 12},
                ),
                row=row,
                col=col,
            )

    def __line_style_to_color_dash(self, style):
        """
        Convert mpl format color-style to plotly format
        :param style: style in format 'color style'. Example: 'r --' means red color and dash style line.
        :return style, color:
        """
        plotly_styles = ["dash", "dot", "dashdot", "solid"]
        splitted = style.split(" ")
        if len(splitted) > 1:
            mpl_color, mpl_style = splitted[0], splitted[1]
        else:
            mpl_color, mpl_style = None, splitted[0]

        plotly_style, plotly_color = None, None

        if mpl_style in plotly_styles:  # specified plotly style line
            plotly_style = mpl_style
        elif mpl_style == ":":
            plotly_style = "dot"
        elif mpl_style == "--":
            plotly_style = "dash"
        elif mpl_style == "-.":
            plotly_style = "dashdot"
        elif mpl_style == "-":
            plotly_style = "solid"

        if plotly_style is None and mpl_color is None:  # it looks only color is specified
            mpl_color = mpl_style

        if mpl_color == "r":
            plotly_color = "red"
        elif mpl_color == "g":
            plotly_color = "green"
        elif mpl_color == "w":
            plotly_color = "white"
        else:
            plotly_color = mpl_color

        return plotly_style, plotly_color

    # def __mpl_color_to_plotly(self, color):
    #     if color == "r":
    #         color = "red"
    #     elif color == "g":
    #         color = "green"
    #     elif color == "w":
    #         color = "white"
    #     return color

    def _show_plot(self, vert_bar, title, zoom):
        # plot all master series
        master_fraction = self.mph / (self.mph + self.sph * len(self.s)) if len(self.s) else 1.0
        row_heights = [master_fraction]
        axis_rules = {}
        if len(self.s):
            row_heights.extend([(1 - master_fraction) / len(self.s)] * len(self.s))

        self.fig = make_subplots(
            rows=len(self.s) + 1, cols=1, shared_xaxes=True, vertical_spacing=0.01, row_heights=row_heights
        )

        ms = self.m if isinstance(self.m, (tuple, list)) else [self.m]
        plot_style = "line"
        for j, m in enumerate(ms):
            if isinstance(m, str):
                if any(
                    [
                        m.startswith(x)
                        for x in [
                            "line",
                            "bar",
                            "step",
                            "stem",
                            "area",
                            "dots",
                            "point",
                            "arrow-up",
                            "arrow-down",
                        ]
                    ]
                ):
                    plot_style = m
                else:
                    self._n_style = m
            else:
                self.__plt_series(m, zoom, "Master", j, 1, 1, plot_style=plot_style)
                self._n_style = "-"

        if vert_bar:
            self.__add_vline("x", vert_bar)
        i = 1
        for k, vs in self.s.items():
            wait_for_limits = False
            i += 1
            vs = vs if isinstance(vs, (tuple, list)) else [vs]
            plot_style = "line"
            self._n_style = "-"
            for j, v in enumerate(vs):
                if wait_for_limits and isinstance(v, (list, tuple)) and len(v) > 1:
                    axis_rules["yaxis%d" % i] = {"range": v}
                    wait_for_limits = False

                elif isinstance(v, str):
                    vl = v.lower()
                    if vl.startswith("lim"):
                        wait_for_limits = True
                        continue
                    elif any(
                        [
                            vl.startswith(x)
                            for x in [
                                "line",
                                "bar",
                                "step",
                                "stem",
                                "area",
                                "dots",
                                "point",
                                "arrow-up",
                                "arrow-down",
                            ]
                        ]
                    ):
                        plot_style = vl
                    else:
                        self._n_style = v
                elif isinstance(v, (float, int)):
                    self.__add_hline("y%d" % i, v)
                else:
                    self.__plt_series(v, zoom, k, j, i, 1, plot_style=plot_style)

        plot_title = title or self._title or ""
        plot_title += str(vert_bar) if vert_bar else ""  # add title to plot
        if plot_title:
            self.fig.update_layout(
                title={"text": plot_title, "y": 0.9, "x": 0.5, "xanchor": "center", "yanchor": "top"}
            )

        self.fig.update_layout(axis_rules)
        self.fig.update_layout(
            xaxis_rangeslider_visible=False, margin=dict(l=5, r=5, t=35, b=5), height=self.mph + self.sph * len(self.s)
        )

        return go.FigureWidget(self.fig)

    def __add_vline(self, xref, x):
        # Line Vertical
        self.fig.add_shape(
            go.layout.Shape(
                type="line", x0=x, x1=x, xref=xref, yref="paper", y0=0, y1=1, line=dict(width=1, dash="dot")
            )
        )

    def __add_hline(self, yref, y):
        # Line Horizontal
        style, color = self.__line_style_to_color_dash(self._n_style)
        self.fig.add_shape(
            go.layout.Shape(
                type="line",
                x0=0,
                x1=1,
                xref="paper",
                yref=yref,
                y0=y,
                y1=y,
                line=dict(width=1, dash=style, color=color),
            )
        )
